# Copyright (C) 2006-2008 Jelmer Vernooij <jelmer@jelmer.uk>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2.1 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA

"""Tests for subvertpy.ra_svn."""

from io import StringIO

from subvertpy.marshall import literal, marshall
from subvertpy.ra_svn import (
    CAPABILITIES,
    MAX_VERSION,
    MECHANISMS,
    MIN_VERSION,
    SVN_PORT,
    SVNConnection,
    SVNServer,
    Editor,
    DirectoryEditor,
    FileEditor,
    Reporter,
    SSHSubprocess,
    SSHVendor,
    mark_busy,
    unmarshall_dirent,
)
from subvertpy.tests import TestCase


class MarkBusyTests(TestCase):
    def test_sets_and_clears_busy(self):
        class Obj:
            busy = False

            @mark_busy
            def do_thing(self):
                return self.busy

        obj = Obj()
        self.assertFalse(obj.busy)
        result = obj.do_thing()
        self.assertTrue(result)  # was True during execution
        self.assertFalse(obj.busy)  # cleared after

    def test_clears_busy_on_exception(self):
        class Obj:
            busy = False

            @mark_busy
            def do_fail(self):
                raise ValueError("fail")

        obj = Obj()
        self.assertRaises(ValueError, obj.do_fail)
        self.assertFalse(obj.busy)

    def test_preserves_docstring(self):
        class Obj:
            @mark_busy
            def do_thing(self):
                """My docstring."""
                pass

        self.assertEqual("My docstring.", Obj.do_thing.__doc__)

    def test_preserves_name(self):
        class Obj:
            @mark_busy
            def do_thing(self):
                pass

        self.assertEqual("do_thing", Obj.do_thing.__name__)

    def test_returns_value(self):
        class Obj:
            busy = False

            @mark_busy
            def do_thing(self):
                return 42

        obj = Obj()
        self.assertEqual(42, obj.do_thing())


class UnmarshallDirentTests(TestCase):
    def test_basic(self):
        d = ["file.txt", "file", 1234, True, 5, [], []]
        result = unmarshall_dirent(d)
        self.assertEqual("file.txt", result["name"])
        self.assertEqual("file", result["kind"])
        self.assertEqual(1234, result["size"])
        self.assertTrue(result["has-props"])
        self.assertEqual(5, result["created-rev"])
        self.assertNotIn("created-date", result)
        self.assertNotIn("last-author", result)

    def test_with_date_and_author(self):
        d = ["dir", "dir", 0, False, 3, "2024-01-01T00:00:00.000000Z", "admin"]
        result = unmarshall_dirent(d)
        self.assertEqual("2024-01-01T00:00:00.000000Z", result["created-date"])
        self.assertEqual("admin", result["last-author"])

    def test_has_props_converted_to_bool(self):
        d = ["f", "file", 0, 0, 1, [], []]
        result = unmarshall_dirent(d)
        self.assertFalse(result["has-props"])

        d2 = ["f", "file", 0, 1, 1, [], []]
        result2 = unmarshall_dirent(d2)
        self.assertTrue(result2["has-props"])


class SVNConnectionTests(TestCase):
    def test_send_msg(self):
        sent = []

        def send_fn(data):
            sent.append(data)

        conn = SVNConnection(None, send_fn)
        conn.send_msg([1, 2, 3])
        self.assertEqual(1, len(sent))
        self.assertEqual(marshall([1, 2, 3]), sent[0])

    def test_send_success(self):
        sent = []

        def send_fn(data):
            sent.append(data)

        conn = SVNConnection(None, send_fn)
        conn.send_success(1, 2)
        self.assertEqual(1, len(sent))
        expected = marshall([literal("success"), [1, 2]])
        self.assertEqual(expected, sent[0])

    def test_send_success_no_args(self):
        sent = []

        def send_fn(data):
            sent.append(data)

        conn = SVNConnection(None, send_fn)
        conn.send_success()
        expected = marshall([literal("success"), []])
        self.assertEqual(expected, sent[0])

    def test_inbuffer_initialized(self):
        conn = SVNConnection(None, None)
        self.assertEqual("", conn.inbuffer)


class EditorTests(TestCase):
    def setUp(self):
        super(EditorTests, self).setUp()
        self.sent = []

        class MockConn:
            _open_ids = []

            def send_msg(_, data):
                self.sent.append(data)

        self.conn = MockConn()

    def test_set_target_revision(self):
        editor = Editor(self.conn)
        editor.set_target_revision(5)
        self.assertEqual(1, len(self.sent))
        self.assertEqual(literal("target-rev"), self.sent[0][0])
        self.assertEqual([5], self.sent[0][1])

    def test_close(self):
        editor = Editor(self.conn)
        editor.close()
        self.assertEqual(literal("close-edit"), self.sent[0][0])

    def test_abort(self):
        editor = Editor(self.conn)
        editor.abort()
        self.assertEqual(literal("abort-edit"), self.sent[0][0])

    def test_open_root_returns_directory_editor(self):
        editor = Editor(self.conn)
        de = editor.open_root()
        self.assertIsInstance(de, DirectoryEditor)

    def test_open_root_no_base_revision(self):
        editor = Editor(self.conn)
        editor.open_root()
        # baserev should be empty list when no base_revision
        self.assertEqual(literal("open-root"), self.sent[0][0])
        self.assertEqual([], self.sent[0][1][0])

    def test_open_root_with_base_revision(self):
        editor = Editor(self.conn)
        editor.open_root(base_revision=5)
        self.assertEqual(literal("open-root"), self.sent[0][0])
        self.assertEqual([5], self.sent[0][1][0])


class DirectoryEditorTests(TestCase):
    def setUp(self):
        super(DirectoryEditorTests, self).setUp()
        self.sent = []

        class MockConn:
            _open_ids = []

            def send_msg(_, data):
                self.sent.append(data)

        self.conn = MockConn()
        self.dir_id = "test-id"
        self.conn._open_ids.append(self.dir_id)
        self.de = DirectoryEditor(self.conn, self.dir_id)

    def test_change_prop(self):
        self.de.change_prop("svn:log", "value")
        self.assertEqual(literal("change-dir-prop"), self.sent[0][0])
        self.assertEqual([self.dir_id, "svn:log", ["value"]], self.sent[0][1])

    def test_change_prop_none_value(self):
        self.de.change_prop("svn:log", None)
        self.assertEqual([self.dir_id, "svn:log", []], self.sent[0][1])

    def test_add_file_returns_file_editor(self):
        fe = self.de.add_file("test.txt")
        self.assertIsInstance(fe, FileEditor)
        self.assertEqual(literal("add-file"), self.sent[0][0])

    def test_add_file_with_copyfrom(self):
        self.de.add_file("test.txt", copyfrom_path="/trunk/old.txt", copyfrom_rev=3)
        args = self.sent[0][1]
        self.assertEqual("test.txt", args[0])
        self.assertEqual(["/trunk/old.txt", 3], args[3])

    def test_add_file_no_copyfrom(self):
        self.de.add_file("test.txt")
        args = self.sent[0][1]
        self.assertEqual([], args[3])  # empty copyfrom_data

    def test_add_directory_returns_directory_editor(self):
        de2 = self.de.add_directory("subdir")
        self.assertIsInstance(de2, DirectoryEditor)
        self.assertEqual(literal("add-dir"), self.sent[0][0])

    def test_add_directory_with_copyfrom(self):
        self.de.add_directory("subdir", copyfrom_path="/trunk/old", copyfrom_rev=2)
        args = self.sent[0][1]
        self.assertEqual("subdir", args[0])
        self.assertEqual(["/trunk/old", 2], args[3])

    def test_add_directory_no_copyfrom(self):
        self.de.add_directory("subdir")
        args = self.sent[0][1]
        self.assertEqual([], args[3])

    def test_open_file(self):
        fe = self.de.open_file("file.txt", 3)
        self.assertIsInstance(fe, FileEditor)
        self.assertEqual(literal("open-file"), self.sent[0][0])
        args = self.sent[0][1]
        self.assertEqual("file.txt", args[0])
        self.assertEqual(3, args[3])

    def test_open_directory(self):
        de2 = self.de.open_directory("subdir", 2)
        self.assertIsInstance(de2, DirectoryEditor)
        self.assertEqual(literal("open-dir"), self.sent[0][0])
        args = self.sent[0][1]
        self.assertEqual("subdir", args[0])
        self.assertEqual(2, args[3])

    def test_delete_entry(self):
        self.de.delete_entry("file.txt", 3)
        self.assertEqual(literal("delete-entry"), self.sent[0][0])

    def test_close(self):
        # DirectoryEditor.__init__ appends the id to _open_ids,
        # and close() pops the last one
        initial_count = self.conn._open_ids.count(self.dir_id)
        self.de.close()
        self.assertEqual(literal("close-dir"), self.sent[0][0])
        self.assertEqual(initial_count - 1, self.conn._open_ids.count(self.dir_id))


class FileEditorTests(TestCase):
    def setUp(self):
        super(FileEditorTests, self).setUp()
        self.sent = []

        class MockConn:
            _open_ids = []

            def send_msg(_, data):
                self.sent.append(data)

        self.conn = MockConn()
        self.file_id = "file-id"
        self.conn._open_ids.append(self.file_id)
        self.fe = FileEditor(self.conn, self.file_id)

    def test_change_prop(self):
        self.fe.change_prop("svn:mime-type", "text/plain")
        self.assertEqual(literal("change-file-prop"), self.sent[0][0])
        self.assertEqual(
            [self.file_id, "svn:mime-type", ["text/plain"]], self.sent[0][1]
        )

    def test_change_prop_none(self):
        self.fe.change_prop("svn:mime-type", None)
        self.assertEqual([self.file_id, "svn:mime-type", []], self.sent[0][1])

    def test_close_no_checksum(self):
        self.fe.close()
        self.assertEqual(literal("close-file"), self.sent[0][0])
        self.assertEqual([self.file_id, []], self.sent[0][1])

    def test_close_with_checksum(self):
        self.fe.close(checksum="abc123")
        self.assertEqual([self.file_id, ["abc123"]], self.sent[0][1])

    def test_apply_textdelta(self):
        handler = self.fe.apply_textdelta()
        self.assertTrue(callable(handler))
        # Should have sent apply-textdelta and the SVN header chunk
        self.assertEqual(literal("apply-textdelta"), self.sent[0][0])
        self.assertEqual(literal("textdelta-chunk"), self.sent[1][0])

    def test_apply_textdelta_with_checksum(self):
        self.fe.apply_textdelta(base_checksum="md5sum")
        self.assertEqual([self.file_id, ["md5sum"]], self.sent[0][1])

    def test_apply_textdelta_handler_sends_chunks(self):
        handler = self.fe.apply_textdelta()
        self.sent.clear()
        # Send a delta window (None signals end)
        handler(None)
        self.assertEqual(literal("textdelta-end"), self.sent[0][0])

    def test_apply_textdelta_handler_sends_window(self):
        from subvertpy.delta import TXDELTA_NEW

        handler = self.fe.apply_textdelta()
        self.sent.clear()
        window = (0, 0, 3, 1, [(TXDELTA_NEW, 0, 3)], b"foo")
        handler(window)
        self.assertEqual(literal("textdelta-chunk"), self.sent[0][0])
        # The second element should be the file_id and packed window data
        self.assertEqual(self.file_id, self.sent[0][1][0])


class SSHSubprocessTests(TestCase):
    def test_get_filelike_channels(self):
        class FakeProc:
            stdout = "fake_stdout"
            stdin = "fake_stdin"

        proc = FakeProc()
        ssh = SSHSubprocess(proc)
        stdout, stdin = ssh.get_filelike_channels()
        self.assertEqual("fake_stdout", stdout)
        self.assertEqual("fake_stdin", stdin)


class SSHVendorTests(TestCase):
    def test_is_instantiable(self):
        vendor = SSHVendor()
        self.assertTrue(hasattr(vendor, "connect_ssh"))


class ReporterTests(TestCase):
    def setUp(self):
        super(ReporterTests, self).setUp()
        self.sent = []

        class MockConn:
            busy = True

            def send_msg(_, data):
                self.sent.append(data)

        self.conn = MockConn()

    def test_set_path_basic(self):
        reporter = Reporter(self.conn, None)
        reporter.set_path("path", 5)
        self.assertEqual(literal("set-path"), self.sent[0][0])
        args = self.sent[0][1]
        self.assertEqual("path", args[0])
        self.assertEqual(5, args[1])
        self.assertFalse(args[2])  # start_empty default
        self.assertEqual([], args[3])  # no lock_token

    def test_set_path_start_empty(self):
        reporter = Reporter(self.conn, None)
        reporter.set_path("path", 5, start_empty=True)
        args = self.sent[0][1]
        self.assertTrue(args[2])

    def test_set_path_with_lock_token(self):
        reporter = Reporter(self.conn, None)
        reporter.set_path("path", 5, lock_token="token123")
        args = self.sent[0][1]
        self.assertEqual(["token123"], args[3])

    def test_set_path_with_depth(self):
        reporter = Reporter(self.conn, None)
        reporter.set_path("path", 5, depth="infinity")
        args = self.sent[0][1]
        self.assertEqual("infinity", args[4])

    def test_delete_path(self):
        reporter = Reporter(self.conn, None)
        reporter.delete_path("deleted/path")
        self.assertEqual(literal("delete-path"), self.sent[0][0])
        self.assertEqual(["deleted/path"], self.sent[0][1])

    def test_link_path_basic(self):
        reporter = Reporter(self.conn, None)
        reporter.link_path("path", "svn://example.com/repo", 3)
        self.assertEqual(literal("link-path"), self.sent[0][0])
        args = self.sent[0][1]
        self.assertEqual("path", args[0])
        self.assertEqual("svn://example.com/repo", args[1])
        self.assertEqual(3, args[2])
        self.assertFalse(args[3])  # start_empty default
        self.assertEqual([], args[4])  # no lock_token

    def test_link_path_with_lock_token(self):
        reporter = Reporter(self.conn, None)
        reporter.link_path("path", "svn://example.com", 3, lock_token="tok")
        args = self.sent[0][1]
        self.assertEqual(["tok"], args[4])

    def test_link_path_with_depth(self):
        reporter = Reporter(self.conn, None)
        reporter.link_path("path", "svn://example.com", 3, depth="files")
        args = self.sent[0][1]
        self.assertEqual("files", args[5])

    def test_abort(self):
        reporter = Reporter(self.conn, None)
        reporter.abort()
        self.assertEqual(literal("abort-report"), self.sent[0][0])
        self.assertFalse(self.conn.busy)


class SVNServerMutterTests(TestCase):
    def test_mutter_with_logf(self):
        logf = StringIO()
        sent = []

        def send_fn(data):
            sent.append(data)

        # Create SVNServer without full init (which sends greeting)
        server = SVNServer.__new__(SVNServer)
        server._logf = logf
        server._stop = False
        server.mutter("hello world")
        self.assertEqual("hello world\n", logf.getvalue())

    def test_mutter_without_logf(self):
        server = SVNServer.__new__(SVNServer)
        server._logf = None
        server._stop = False
        # Should not raise
        server.mutter("hello world")

    def test_close_sets_stop(self):
        server = SVNServer.__new__(SVNServer)
        server._stop = False
        server.close()
        self.assertTrue(server._stop)

    def test_send_failure(self):
        sent = []

        def send_fn(data):
            sent.append(data)

        server = SVNServer.__new__(SVNServer)
        server.send_fn = send_fn
        server.inbuffer = ""
        server.recv_fn = None
        server.send_failure([210001, "Unknown command", "file.py", 1])
        self.assertEqual(1, len(sent))

    def test_send_ack(self):
        sent = []

        def send_fn(data):
            sent.append(data)

        server = SVNServer.__new__(SVNServer)
        server.send_fn = send_fn
        server.inbuffer = ""
        server.recv_fn = None
        server.send_ack()
        self.assertEqual(1, len(sent))

    def test_send_unknown(self):
        sent = []

        def send_fn(data):
            sent.append(data)

        server = SVNServer.__new__(SVNServer)
        server.send_fn = send_fn
        server.inbuffer = ""
        server.recv_fn = None
        server.send_unknown("bogus-cmd")
        self.assertEqual(1, len(sent))


class ConstantsTests(TestCase):
    def test_svn_port(self):
        self.assertEqual(3690, SVN_PORT)

    def test_min_version(self):
        self.assertEqual(2, MIN_VERSION)

    def test_max_version(self):
        self.assertEqual(2, MAX_VERSION)

    def test_capabilities(self):
        self.assertIsInstance(CAPABILITIES, list)
        self.assertIn("edit-pipeline", CAPABILITIES)

    def test_mechanisms(self):
        self.assertIsInstance(MECHANISMS, list)
        self.assertIn("ANONYMOUS", MECHANISMS)



#!/usr/bin/python
# Demonstrates how to do access the working tree using subvertpy

import os
from subvertpy import client, repos, wc
from subvertpy.ra import Auth, get_username_provider

# Create a repository
repos.create("tmprepo")

c = client.Client(auth=Auth([get_username_provider()]))
c.checkout("file://" + os.getcwd() + "/tmprepo", "tmpco", "HEAD")

w = wc.WorkingCopy(None, "tmpco")
print(w)
entry = w.entry("tmpco")
print(entry.revision)
print(entry.url)



# Copyright (C) 2005-2007 Jelmer Vernooij <jelmer@jelmer.uk>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2.1 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Subversion core library tests."""

import os
import tempfile
import types

import subvertpy
from subvertpy.tests import TestCase


class TestCore(TestCase):
    def setUp(self):
        super(TestCore, self).setUp()

    def test_exc(self):
        self.assertTrue(isinstance(subvertpy.SubversionException("foo", 1), Exception))

    def test_exc_args(self):
        exc = subvertpy.SubversionException("test message", 42)
        self.assertEqual(exc.args, ("test message", 42))

    def test_exc_child(self):
        child = subvertpy.SubversionException("child", 2)
        exc = subvertpy.SubversionException("parent", 1, child=child)
        self.assertEqual(exc.child, child)

    def test_exc_location(self):
        exc = subvertpy.SubversionException("msg", 1, location="file.c:42")
        self.assertEqual(exc.location, "file.c:42")

    def test_exc_child_default_none(self):
        exc = subvertpy.SubversionException("msg", 1)
        self.assertIsNone(exc.child)

    def test_exc_location_default_none(self):
        exc = subvertpy.SubversionException("msg", 1)
        self.assertIsNone(exc.location)


class TestConstants(TestCase):
    def test_node_constants(self):
        self.assertEqual(subvertpy.NODE_NONE, 0)
        self.assertEqual(subvertpy.NODE_FILE, 1)
        self.assertEqual(subvertpy.NODE_DIR, 2)
        self.assertEqual(subvertpy.NODE_UNKNOWN, 3)

    def test_error_constants_are_ints(self):
        self.assertIsInstance(subvertpy.ERR_UNSUPPORTED_FEATURE, int)
        self.assertIsInstance(subvertpy.ERR_RA_SVN_UNKNOWN_CMD, int)
        self.assertIsInstance(subvertpy.ERR_FS_NO_SUCH_REVISION, int)
        self.assertIsInstance(subvertpy.ERR_CANCELLED, int)

    def test_error_constants_values(self):
        self.assertEqual(subvertpy.ERR_UNSUPPORTED_FEATURE, 200007)
        self.assertEqual(subvertpy.ERR_RA_SVN_UNKNOWN_CMD, 210001)
        self.assertEqual(subvertpy.ERR_RA_SVN_CONNECTION_CLOSED, 210002)
        self.assertEqual(subvertpy.ERR_WC_LOCKED, 155004)
        self.assertEqual(subvertpy.ERR_RA_NOT_AUTHORIZED, 170001)
        self.assertEqual(subvertpy.ERR_INCOMPLETE_DATA, 200003)
        self.assertEqual(subvertpy.ERR_FS_NO_SUCH_REVISION, 160006)
        self.assertEqual(subvertpy.ERR_CANCELLED, 200015)

    def test_wc_not_working_copy_aliases(self):
        self.assertEqual(
            subvertpy.ERR_WC_NOT_WORKING_COPY, subvertpy.ERR_WC_NOT_DIRECTORY
        )

    def test_auth_param_constants(self):
        self.assertEqual(subvertpy.AUTH_PARAM_DEFAULT_USERNAME, "svn:auth:username")
        self.assertEqual(subvertpy.AUTH_PARAM_DEFAULT_PASSWORD, "svn:auth:password")

    def test_ssl_constants(self):
        self.assertEqual(subvertpy.SSL_NOTYETVALID, 0x00000001)
        self.assertEqual(subvertpy.SSL_EXPIRED, 0x00000002)
        self.assertEqual(subvertpy.SSL_CNMISMATCH, 0x00000004)
        self.assertEqual(subvertpy.SSL_UNKNOWNCA, 0x00000008)
        self.assertEqual(subvertpy.SSL_OTHER, 0x40000000)

    def test_apr_error_constants(self):
        self.assertEqual(subvertpy.ERR_APR_OS_START_EAIERR, 670000)
        self.assertEqual(subvertpy.ERR_APR_OS_ERRSPACE_SIZE, 50000)
        self.assertEqual(subvertpy.ERR_CATEGORY_SIZE, 5000)


class TestCheckMtime(TestCase):
    def test_check_mtime_c_file_missing(self):
        # When no .c file exists, _check_mtime should return True
        module = types.ModuleType("fake")
        module.__file__ = "/nonexistent/path/fake.so"
        self.assertTrue(subvertpy._check_mtime(module))

    def test_check_mtime_c_file_older(self):
        # When .c file exists but is older than the module, return True
        tmpdir = tempfile.mkdtemp()
        try:
            so_path = os.path.join(tmpdir, "mod.so")
            c_path = os.path.join(tmpdir, "mod.c")
            # Create .c first (older)
            with open(c_path, "w") as f:
                f.write("/* c */")
            # Then .so (newer) - ensure different mtime
            os.utime(c_path, (1000, 1000))
            with open(so_path, "w") as f:
                f.write("so")
            os.utime(so_path, (2000, 2000))
            module = types.ModuleType("mod")
            module.__file__ = so_path
            self.assertTrue(subvertpy._check_mtime(module))
        finally:
            import shutil

            shutil.rmtree(tmpdir)

    def test_check_mtime_c_file_newer(self):
        # When .c file exists and is newer than the module, return False
        tmpdir = tempfile.mkdtemp()
        try:
            so_path = os.path.join(tmpdir, "mod.so")
            c_path = os.path.join(tmpdir, "mod.c")
            # Create .so first (older)
            with open(so_path, "w") as f:
                f.write("so")
            os.utime(so_path, (1000, 1000))
            # Then .c (newer)
            with open(c_path, "w") as f:
                f.write("/* c */")
            os.utime(c_path, (2000, 2000))
            module = types.ModuleType("mod")
            module.__file__ = so_path
            self.assertFalse(subvertpy._check_mtime(module))
        finally:
            import shutil

            shutil.rmtree(tmpdir)

    def test_version(self):
        self.assertIsInstance(subvertpy.__version__, tuple)
        self.assertEqual(len(subvertpy.__version__), 3)



# Copyright (C) 2005-2007 Jelmer Vernooij <jelmer@jelmer.uk>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2.1 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Subversion server tests."""

import re

from subvertpy.server import (
    ServerBackend,
    ServerRepositoryBackend,
    generate_random_id,
)
from subvertpy.tests import TestCase


class GenerateRandomIdTests(TestCase):
    def test_returns_string(self):
        result = generate_random_id()
        self.assertIsInstance(result, str)

    def test_is_uuid_format(self):
        result = generate_random_id()
        uuid_re = re.compile(
            r"^[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-"
            r"[0-9a-f]{4}-[0-9a-f]{12}$"
        )
        self.assertRegex(result, uuid_re)

    def test_unique(self):
        id1 = generate_random_id()
        id2 = generate_random_id()
        self.assertNotEqual(id1, id2)


class ServerBackendTests(TestCase):
    def test_open_repository_raises(self):
        backend = ServerBackend()
        self.assertRaises(NotImplementedError, backend.open_repository, "/test")


class ServerRepositoryBackendTests(TestCase):
    def setUp(self):
        super(ServerRepositoryBackendTests, self).setUp()
        self.backend = ServerRepositoryBackend()

    def test_get_uuid_raises(self):
        self.assertRaises(NotImplementedError, self.backend.get_uuid)

    def test_get_latest_revnum_raises(self):
        self.assertRaises(NotImplementedError, self.backend.get_latest_revnum)

    def test_log_raises(self):
        self.assertRaises(
            NotImplementedError, self.backend.log, None, "/", 0, 1, True, True, 0
        )

    def test_update_raises(self):
        self.assertRaises(NotImplementedError, self.backend.update, None, 1, "/", True)

    def test_check_path_raises(self):
        self.assertRaises(NotImplementedError, self.backend.check_path, "/", 0)

    def test_stat_raises(self):
        self.assertRaises(NotImplementedError, self.backend.stat, "/", 0)

    def test_rev_proplist_raises(self):
        self.assertRaises(NotImplementedError, self.backend.rev_proplist, 0)

    def test_get_locations_raises(self):
        self.assertRaises(NotImplementedError, self.backend.get_locations, "/", 0, [0])



#!/usr/bin/python
# Demonstrates how to do a new commit using Subvertpy

import os
from io import BytesIO
from subvertpy import delta, repos
from subvertpy.ra import RemoteAccess, Auth, get_username_provider

# Create a repository
repos.create("tmprepo")

# Connect to the "remote" repository using the file transport.
# Note that a username provider needs to be provided, so that Subversion
# knows who to record as the author of new commits made over this connection.
repo_url = "file://%s" % os.path.abspath("tmprepo")
conn = RemoteAccess(repo_url, auth=Auth([get_username_provider()]))

# Simple commit that adds a directory
editor = conn.get_commit_editor({"svn:log": "Commit message"})
root = editor.open_root()
# Add a directory
dir = root.add_directory("somedir")
dir.close()
# Add and edit a file
file = root.add_file("somefile")
# Set the svn:executable attribute
file.change_prop("svn:executable", "*")
# Obtain a textdelta handler and send the new file contents
txdelta = file.apply_textdelta()
delta.send_stream(BytesIO(b"new file contents"), txdelta)
file.close()
root.close()
editor.close()

# Rename the directory
editor = conn.get_commit_editor({"svn:log": "Commit message"})
root = editor.open_root()
# Create a new directory copied from somedir:1
dir = root.add_directory("new dir name", "%s/somedir" % repo_url, 1)
dir.close()
# Remove the original directory
root.delete_entry("somedir")
root.close()
editor.close()



# Copyright (C) 2006-2008 Jelmer Vernooij <jelmer@jelmer.uk>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2.1 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA

"""Tests for subvertpy."""

__author__ = "Jelmer Vernooij <jelmer@jelmer.uk>"
__docformat__ = "restructuredText"

from io import BytesIO
import os
import shutil
import stat
import sys
import tempfile
import unittest

try:
    import urlparse
except ImportError:
    import urllib.parse as urlparse
try:
    from urllib import pathname2url
except ImportError:
    from urllib.request import pathname2url

from subvertpy import (
    client,
    delta,
    properties,
    ra,
    repos,
)
from subvertpy.ra import (
    Auth,
    RemoteAccess,
)


def rmtree_with_readonly(path):
    """Simple wrapper for shutil.rmtree that can remove read-only files.

    In Windows a read-only file cannot be removed, and shutil.rmtree fails.
    """

    def force_rm_handle(remove_path, path, excinfo):
        os.chmod(
            path, os.stat(path).st_mode | stat.S_IWUSR | stat.S_IWGRP | stat.S_IWOTH
        )
        remove_path(path)

    shutil.rmtree(path, onerror=force_rm_handle)


class TestCase(unittest.TestCase):
    """Base test case.

    :note: Adds assertIsInstance and assertIs.
    """

    def assertIsInstance(self, obj, kls, msg=None):
        """Fail if obj is not an instance of kls"""
        if not isinstance(obj, kls):
            if msg is None:
                msg = "%r is an instance of %s rather than %s" % (
                    obj,
                    obj.__class__,
                    kls,
                )
            self.fail(msg)

    def assertIs(self, left, right, message=None):
        if left is not right:
            if message is not None:
                raise AssertionError(message)
            else:
                raise AssertionError("%r is not %r." % (left, right))


class TestCaseInTempDir(TestCase):
    """Test case that runs in a temporary directory."""

    def setUp(self):
        TestCase.setUp(self)
        self._oldcwd = os.getcwd()
        self.test_dir = tempfile.mkdtemp()
        os.chdir(self.test_dir)

    def tearDown(self):
        TestCase.tearDown(self)
        os.chdir(self._oldcwd)
        rmtree_with_readonly(self.test_dir)


class TestFileEditor(object):
    """Simple file editor wrapper that doesn't require closing."""

    def __init__(self, file):
        self.file = file
        self.is_closed = False

    def change_prop(self, name, value):
        self.file.change_prop(name, value)

    def modify(self, contents=None):
        if contents is None:
            contents = os.urandom(100)
        txdelta = self.file.apply_textdelta()
        delta.send_stream(BytesIO(contents), txdelta)

    def close(self):
        assert not self.is_closed
        self.is_closed = True
        self.file.close()


class TestDirEditor(object):
    """Simple dir editor wrapper that doesn't require closing."""

    def __init__(self, dir, baseurl, revnum):
        self.dir = dir
        self.baseurl = baseurl
        self.revnum = revnum
        self.is_closed = False
        self.children = []

    def close_children(self):
        for c in reversed(self.children):
            if not c.is_closed:
                c.close()

    def close(self):
        assert not self.is_closed
        self.is_closed = True
        self.close_children()
        self.dir.close()

    def change_prop(self, name, value):
        self.close_children()
        self.dir.change_prop(name, value)

    def open_dir(self, path):
        self.close_children()
        child = TestDirEditor(
            self.dir.open_directory(path, -1), self.baseurl, self.revnum
        )
        self.children.append(child)
        return child

    def open_file(self, path):
        self.close_children()
        child = TestFileEditor(self.dir.open_file(path, -1))
        self.children.append(child)
        return child

    def add_dir(self, path, copyfrom_path=None, copyfrom_rev=-1):
        self.close_children()
        if copyfrom_path is not None:
            copyfrom_path = urlparse.urljoin(self.baseurl + "/", copyfrom_path)
        if copyfrom_path is not None and copyfrom_rev == -1:
            copyfrom_rev = self.revnum
        assert (copyfrom_path is None and copyfrom_rev == -1) or (
            copyfrom_path is not None and copyfrom_rev > -1
        )
        child = TestDirEditor(
            self.dir.add_directory(path, copyfrom_path, copyfrom_rev),
            self.baseurl,
            self.revnum,
        )
        self.children.append(child)
        return child

    def add_file(self, path, copyfrom_path=None, copyfrom_rev=-1):
        self.close_children()
        if copyfrom_path is not None:
            copyfrom_path = urlparse.urljoin(self.baseurl + "/", copyfrom_path)
        if copyfrom_path is not None and copyfrom_rev == -1:
            copyfrom_rev = self.revnum
        child = TestFileEditor(self.dir.add_file(path, copyfrom_path, copyfrom_rev))
        self.children.append(child)
        return child

    def delete(self, path):
        self.close_children()
        self.dir.delete_entry(path)


class TestCommitEditor(TestDirEditor):
    """Simple commit editor wrapper."""

    def __init__(self, editor, baseurl, revnum):
        self.editor = editor
        TestDirEditor.__init__(self, self.editor.open_root(), baseurl, revnum)

    def close(self):
        TestDirEditor.close(self)
        self.editor.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        self.close()
        return False


class SubversionTestCase(TestCaseInTempDir):
    """A test case that provides the ability to build Subversion
    repositories."""

    def _init_client(self):
        self.client_ctx = client.Client()
        self.client_ctx.auth = Auth(
            [
                ra.get_simple_provider(),
                ra.get_username_provider(),
                ra.get_ssl_client_cert_file_provider(),
                ra.get_ssl_client_cert_pw_file_provider(),
                ra.get_ssl_server_trust_file_provider(),
            ]
        )
        self.client_ctx.log_msg_func = self.log_message_func
        # self.client_ctx.notify_func = lambda err: mutter("Error: %s" % err)

    def setUp(self):
        super(SubversionTestCase, self).setUp()
        self._init_client()

    def tearDown(self):
        del self.client_ctx
        super(SubversionTestCase, self).tearDown()

    def log_message_func(self, items):
        return self.next_message

    def make_repository(self, relpath, allow_revprop_changes=True):
        """Create a repository.

        :return: Handle to the repository.
        """
        abspath = os.path.join(self.test_dir, relpath)

        repos.create(abspath)

        if allow_revprop_changes:
            if sys.platform == "win32":
                revprop_hook = os.path.join(abspath, "hooks", "pre-revprop-change.bat")
                with open(revprop_hook, "w") as f:
                    f.write("exit 0\n")
            else:
                revprop_hook = os.path.join(abspath, "hooks", "pre-revprop-change")
                with open(revprop_hook, "w") as f:
                    f.write("#!/bin/sh\n")
                os.chmod(revprop_hook, os.stat(revprop_hook).st_mode | 0o111)

        if sys.platform == "win32":
            return "file:%s" % pathname2url(abspath)
        else:
            return "file://%s" % abspath

    def make_checkout(self, repos_url, relpath):
        """Create a new checkout."""
        self.client_ctx.checkout(repos_url, relpath, "HEAD")

    def client_set_prop(self, path, name, value):
        """Set a property on a local file or directory."""
        if value is None:
            value = ""
        self.client_ctx.propset(name, value, path, False, True)

    def client_get_prop(self, path, name, revnum=None, recursive=False):
        """Retrieve a property from a local or remote file or directory."""
        if revnum is None:
            rev = "WORKING"
        else:
            rev = revnum
        ret = self.client_ctx.propget(name, path, rev, rev, recursive)
        if recursive:
            return ret
        else:
            return next(iter(ret.values()))

    def client_get_revprop(self, url, revnum, name):
        """Get the revision property.

        :param url: URL of the repository
        :param revnum: Revision number
        :param name: Property name
        :return: Revision property value
        """
        r = ra.RemoteAccess(url)
        return r.rev_proplist(revnum)[name]

    def client_set_revprop(self, url, revnum, name, value):
        """Set a revision property on a repository.

        :param url: URL of the repository
        :param revnum: Revision number of the revision
        :param name: Name of the property
        :param value: Value of the property, None to remove
        """
        r = ra.RemoteAccess(url, auth=Auth([ra.get_username_provider()]))
        r.change_rev_prop(revnum, name, value)

    def client_resolve(self, path, choice, depth=0):
        """Resolve a conflict set on a local path."""
        self.client_ctx.resolve(path, depth, choice)

    def client_lock(self, path, comment="A comment", steal_lock=False):
        self.client_ctx.lock(path, comment, steal_lock)

    def client_unlock(self, path, steal_lock=False):
        self.client_ctx.unlock(path, steal_lock)

    def client_commit(self, dir, message=None, recursive=True):
        """Commit current changes in specified working copy.

        :param dir: List of paths to commit.
        """
        olddir = os.path.abspath(".")
        self.next_message = message
        os.chdir(dir)
        info = []

        def add_info(*args):
            info.append(args)

        self.client_ctx.commit(["."], recursive, False, callback=add_info)
        os.chdir(olddir)
        assert len(info) == 1
        return info[0]

    def client_add(self, relpath, recursive=True):
        """Add specified files to working copy.

        :param relpath: Path to the files to add.
        """
        self.client_ctx.add(relpath, recursive, False, False)

    def client_log(self, url, start_revnum, stop_revnum):
        """Fetch the log

        :param url: URL to log
        :param start_revnum: Start revision of the range to log over
        :param start_revnum: Stop revision of the range to log over
        :return: Dictionary
        """
        r = ra.RemoteAccess(url)
        ret = {}

        def rcvr(orig_paths, rev, revprops, has_children=None):
            ret[rev] = (
                orig_paths,
                revprops.get(properties.PROP_REVISION_AUTHOR),
                revprops.get(properties.PROP_REVISION_DATE),
                revprops.get(properties.PROP_REVISION_LOG),
            )

        r.get_log(
            rcvr,
            [""],
            start_revnum,
            stop_revnum,
            0,
            True,
            True,
            revprops=[
                properties.PROP_REVISION_AUTHOR,
                properties.PROP_REVISION_DATE,
                properties.PROP_REVISION_LOG,
            ],
        )
        return ret

    def client_delete(self, relpath):
        """Remove specified files from working copy.

        :param relpath: Path to the files to remove.
        """
        self.client_ctx.delete([relpath], True)

    def client_copy(self, oldpath, newpath, revnum=None):
        """Copy file in working copy.

        :param oldpath: Relative path to original file.
        :param newpath: Relative path to new file.
        """
        if revnum is None:
            rev = "HEAD"
        else:
            rev = revnum
        self.client_ctx.copy(oldpath, newpath, rev)

    def client_update(self, path):
        """Update path.

        :param path: Path
        """
        self.client_ctx.update([path], "HEAD", True)

    def build_tree(self, files):
        """Create a directory tree.

        :param files: Dictionary with filenames as keys, contents as
            values. None as value indicates a directory.
        """
        for name, content in files.items():
            if content is None:
                try:
                    os.makedirs(name)
                except OSError:
                    pass
            else:
                try:
                    os.makedirs(os.path.dirname(name))
                except OSError:
                    pass
                with open(name, "wb") as f:
                    f.write(content)

    def make_client(self, repospath, clientpath, allow_revprop_changes=True):
        """Create a repository and a checkout. Return the checkout.

        :param repospath: Optional relpath to check out if not the full
            repository.
        :param clientpath: Path to checkout
        :return: Repository URL.
        """
        repos_url = self.make_repository(
            repospath, allow_revprop_changes=allow_revprop_changes
        )
        self.make_checkout(repos_url, clientpath)
        return repos_url

    def open_fs(self, relpath):
        """Open a fs.

        :return: FS.
        """
        return repos.Repository(relpath).fs()

    def get_commit_editor(self, url, message="Test commit"):
        """Obtain a commit editor.

        :param url: URL to connect to
        :param message: Commit message
        :return: Commit editor object
        """
        ra_ctx = RemoteAccess(
            url.encode("utf-8"), auth=Auth([ra.get_username_provider()])
        )
        revnum = ra_ctx.get_latest_revnum()
        return TestCommitEditor(
            ra_ctx.get_commit_editor({"svn:log": message}), ra_ctx.url, revnum
        )


def test_suite():
    names = [
        "client",
        "core",
        "delta",
        "marshall",
        "properties",
        "ra",
        "ra_svn",
        "repos",
        "server",
        "subr",
        "wc",
    ]
    module_names = ["subvertpy.tests.test_" + name for name in names]
    result = unittest.TestSuite()
    loader = unittest.TestLoader()
    suite = loader.loadTestsFromNames(module_names)
    result.addTests(suite)
    return result



# Copyright (C) 2005-2006 Jelmer Vernooij <jelmer@jelmer.uk>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2.1 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
"""Subversion delta operations."""

__author__ = "Jelmer Vernooij <jelmer@jelmer.uk>"
__docformat__ = "restructuredText"

from hashlib import (
    md5,
)


TXDELTA_SOURCE = 0
TXDELTA_TARGET = 1
TXDELTA_NEW = 2
TXDELTA_INVALID = 3

MAX_ENCODED_INT_LEN = 10

DELTA_WINDOW_SIZE = 102400


def apply_txdelta_window(sbuf, window):
    """Apply a txdelta window to a buffer.

    :param sbuf: Source buffer (as bytestring)
    :param window: (sview_offset, sview_len, tview_len, src_ops, ops, new_data)
    :param sview_offset: Offset of the source view
    :param sview_len: Length of the source view
    :param tview_len: Target view length
    :param src_ops: Operations to apply to sview
    :param ops: Ops to apply
    :param new_data: Buffer with possible new data
    :return: Target buffer
    """
    (sview_offset, sview_len, tview_len, src_ops, ops, new_data) = window
    sview = sbuf[sview_offset : sview_offset + sview_len]
    tview = txdelta_apply_ops(src_ops, ops, new_data, sview)
    if len(tview) != tview_len:
        raise AssertionError("%d != %d" % (len(tview), tview_len))
    return tview


def apply_txdelta_handler_chunks(source_chunks, target_chunks):
    """Return a function that can be called repeatedly with txdelta windows.

    :param sbuf: Source buffer
    :param target_stream: Target stream
    """
    sbuf = bytes().join(source_chunks)

    def apply_window(window):
        if window is None:
            return  # Last call
        target_chunks.append(apply_txdelta_window(sbuf, window))

    return apply_window


def apply_txdelta_handler(sbuf, target_stream):
    """Return a function that can be called repeatedly with txdelta windows.

    :param sbuf: Source buffer
    :param target_stream: Target stream
    """

    def apply_window(window):
        if window is None:
            return  # Last call
        target_stream.write(apply_txdelta_window(sbuf, window))

    return apply_window


def txdelta_apply_ops(src_ops, ops, new_data, sview):
    """Apply txdelta operations to a source view.

    :param src_ops: Source operations, ignored.
    :param ops: List of operations (action, offset, length).
    :param new_data: Buffer to fetch fragments with new data from
    :param sview: Source data
    :return: Result data
    """
    tview = bytearray()
    for action, offset, length in ops:
        if action == TXDELTA_SOURCE:
            # Copy from source area.
            tview.extend(sview[offset : offset + length])
        elif action == TXDELTA_TARGET:
            for i in range(length):
                tview.append(tview[offset + i])
        elif action == TXDELTA_NEW:
            tview.extend(new_data[offset : offset + length])
        else:
            raise Exception("Invalid delta instruction code")
    return tview


def send_stream(stream, handler, block_size=DELTA_WINDOW_SIZE):
    """Send txdelta windows that create stream to handler

    :param stream: file-like object to read the file from
    :param handler: txdelta window handler function
    :return: MD5 hash over the stream
    """
    hash = md5()
    text = stream.read(block_size)
    if not isinstance(text, bytes):
        raise TypeError("The stream should read out bytes")
    while text:
        hash.update(text)
        window = (0, 0, len(text), 0, [(TXDELTA_NEW, 0, len(text))], text)
        handler(window)
        text = stream.read(block_size)
    handler(None)
    return hash.digest()


def encode_length(len):
    """Encode a length variable.

    :param len: Length to encode
    :return: String with encoded length
    """
    # Based on encode_int() in subversion/libsvn_delta/svndiff.c
    assert len >= 0
    assert isinstance(len, int), "expected int, got %r" % (len,)

    # Count number of required bytes
    v = len >> 7
    n = 1
    while v > 0:
        v = v >> 7
        n += 1

    assert n <= MAX_ENCODED_INT_LEN

    ret = bytearray()
    while n > 0:
        n -= 1
        if n > 0:
            cont = 1
        else:
            cont = 0
        ret.append(((len >> (n * 7)) & 0x7F) | (cont << 7))

    return ret


def decode_length(text):
    """Decode a length variable.

    :param text: Bytestring to decode
    :return: Integer with actual length
    """
    # Decode bytes until we're done.  */
    ret = 0
    next = True
    while next:
        ret = (ret << 7) | (text[0] & 0x7F)
        next = (text[0] >> 7) & 0x1
        text = text[1:]
    return ret, text


def pack_svndiff_instruction(diff_params):
    """Pack a SVN diff instruction

    :param diff_params: (action, offset, length)
    :param action: Action
    :param offset: Offset
    :param length: Length
    :return: encoded text
    """
    (action, offset, length) = diff_params
    if length < 0x3F:
        text = bytearray(((action << 6) + length,))
    else:
        text = bytearray((action << 6,)) + encode_length(length)
    if action != TXDELTA_NEW:
        text += encode_length(offset)
    return text


def unpack_svndiff_instruction(text):
    """Unpack a SVN diff instruction

    :param text: Text to parse
    :return: tuple with operation, remaining text
    """
    action = text[0] >> 6
    length = text[0] & 0x3F
    text = text[1:]
    assert action in (TXDELTA_NEW, TXDELTA_SOURCE, TXDELTA_TARGET)
    if length == 0:
        length, text = decode_length(text)
    if action != TXDELTA_NEW:
        offset, text = decode_length(text)
    else:
        offset = 0
    return (action, offset, length), text


SVNDIFF0_HEADER = b"SVN\0"


def pack_svndiff0_window(window):
    """Pack an individual window using svndiff0.

    :param window: Window to pack
    :return: Packed diff (as bytestring)
    """
    (sview_offset, sview_len, tview_len, src_ops, ops, new_data) = window
    ret = (
        encode_length(sview_offset)
        + encode_length(sview_len)
        + encode_length(tview_len)
    )

    instrdata = bytearray()
    for op in ops:
        instrdata += pack_svndiff_instruction(op)

    ret.extend(encode_length(len(instrdata)))
    ret.extend(encode_length(len(new_data)))
    ret.extend(instrdata)
    ret.extend(new_data)
    return ret


def pack_svndiff0(windows):
    """Pack a SVN diff file.

    :param windows: Iterator over diff windows
    :return: text
    """
    ret = SVNDIFF0_HEADER
    for window in windows:
        ret += pack_svndiff0_window(window)
    return ret


def unpack_svndiff0(text):
    """Unpack a version 0 svndiff text.

    :param text: Text to unpack.
    :return: yields tuples with sview_offset, sview_len, tview_len, ops_len,
        ops, newdata
    """
    assert text.startswith(SVNDIFF0_HEADER)
    text = text[4:]

    while text:
        sview_offset, text = decode_length(text)
        sview_len, text = decode_length(text)
        tview_len, text = decode_length(text)
        instr_len, text = decode_length(text)
        newdata_len, text = decode_length(text)

        instrdata = text[:instr_len]
        text = text[instr_len:]

        ops = []
        while instrdata:
            op, instrdata = unpack_svndiff_instruction(instrdata)
            ops.append(op)

        newdata = text[:newdata_len]
        text = text[newdata_len:]
        yield (sview_offset, sview_len, tview_len, len(ops), ops, newdata)



#!/usr/bin/python
# Demonstrates how to iterate over the log of a Subversion repository.

from subvertpy.ra import RemoteAccess

conn = RemoteAccess("svn://svn.samba.org/subvertpy/trunk")

for changed_paths, rev, revprops, has_children in conn.iter_log(
    paths=None, start=0, end=conn.get_latest_revnum(), discover_changed_paths=True
):
    print("=" * 79)
    print("%d:" % rev)
    print("Revision properties:")
    for entry in revprops.items():
        print("  %s: %s" % entry)
    print("")

    print("Changed paths")
    for path, (action, from_path, from_rev, node_kind) in changed_paths.items():
        print("  %s (%s)" % (path, action))



# Copyright (C) 2005-2007 Jelmer Vernooij <jelmer@jelmer.uk>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2.1 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Subversion repository library tests."""

from io import BytesIO
import os
import textwrap

from subvertpy import repos, SubversionException
from subvertpy.tests import SubversionTestCase, TestCaseInTempDir, TestCase


class VersionTest(TestCase):
    def test_version_length(self):
        self.assertEqual(4, len(repos.version()))

    def test_api_version_length(self):
        self.assertEqual(4, len(repos.api_version()))

    def test_api_version_later_same(self):
        self.assertTrue(repos.api_version() <= repos.version())


class TestRepository(TestCaseInTempDir):
    def setUp(self):
        super(TestRepository, self).setUp()

    def test_create(self):
        repos.create(os.path.join(self.test_dir, "foo"))

    def test_capability(self):
        r = repos.create(os.path.join(self.test_dir, "foo"))
        self.assertIsInstance(r.has_capability("mergeinfo"), bool)

    def test_verify_fs(self):
        r = repos.create(os.path.join(self.test_dir, "foo"))
        f = BytesIO()
        r.verify_fs(f, 0, 0)
        self.assertEqual(b"* Verified revision 0.\n", f.getvalue())

    def test_open(self):
        repos.create(os.path.join(self.test_dir, "foo"))
        repos.Repository("foo")

    def test_uuid(self):
        repos.create(os.path.join(self.test_dir, "foo"))
        self.assertEqual(36, len(repos.Repository("foo").fs().get_uuid()))

    def test_youngest_rev(self):
        repos.create(os.path.join(self.test_dir, "foo"))
        self.assertEqual(0, repos.Repository("foo").fs().youngest_revision())

    def test_rev_root(self):
        repos.create(os.path.join(self.test_dir, "foo"))
        self.assertTrue(repos.Repository("foo").fs().revision_root(0) is not None)

    def test_load_fs_with_hooks(self):
        r = repos.create(os.path.join(self.test_dir, "hooks_test"))
        dumpfile = textwrap.dedent("""\
        SVN-fs-dump-format-version: 2

        UUID: 38f0a982-fd1f-4e00-aa6b-a20720f4b9ca

        Revision-number: 0
        Prop-content-length: 56
        Content-length: 56

        K 8
        svn:date
        V 27
        2011-08-26T13:08:30.187858Z
        PROPS-END
        """).encode("ascii")
        feedback = BytesIO()
        r.load_fs(
            BytesIO(dumpfile),
            feedback,
            repos.LOAD_UUID_DEFAULT,
            use_pre_commit_hook=False,
            use_post_commit_hook=False,
        )

    def test_hotcopy(self):
        src = os.path.join(self.test_dir, "hotcopy_src")
        dest = os.path.join(self.test_dir, "hotcopy_dest")
        repos.create(src)
        repos.hotcopy(src, dest, True)
        self.assertTrue(os.path.exists(dest))

    def test_load_fs_invalid(self):
        r = repos.create(os.path.join(self.test_dir, "foo"))
        dumpfile = b"Malformed"
        feedback = BytesIO()
        self.assertRaises(
            SubversionException,
            r.load_fs,
            BytesIO(dumpfile),
            feedback,
            repos.LOAD_UUID_DEFAULT,
        )

    def test_load_fs(self):
        r = repos.create(os.path.join(self.test_dir, "foo"))
        dumpfile = textwrap.dedent("""\
        SVN-fs-dump-format-version: 2

        UUID: 38f0a982-fd1f-4e00-aa6b-a20720f4b9ca

        Revision-number: 0
        Prop-content-length: 56
        Content-length: 56

        K 8
        svn:date
        V 27
        2011-08-26T13:08:30.187858Z
        PROPS-END
        """).encode("ascii")
        feedback = BytesIO()
        r.load_fs(BytesIO(dumpfile), feedback, repos.LOAD_UUID_DEFAULT)
        self.assertEqual(r.fs().get_uuid(), "38f0a982-fd1f-4e00-aa6b-a20720f4b9ca")

    def test_rev_props(self):
        repos.create(os.path.join(self.test_dir, "foo"))
        self.assertEqual(
            ["svn:date"], list(repos.Repository("foo").fs().revision_proplist(0).keys())
        )

    def test_rev_root_invalid(self):
        repos.create(os.path.join(self.test_dir, "foo"))
        self.assertRaises(
            SubversionException, repos.Repository("foo").fs().revision_root, 1
        )

    def test_pack_fs(self):
        r = repos.create(os.path.join(self.test_dir, "foo"))
        r.pack_fs()

    def test_pack_fs_with_notify(self):
        r = repos.create(os.path.join(self.test_dir, "pack_notify"))
        notifications = []

        def notify_func(shard, action):
            notifications.append((shard, action))

        r.pack_fs(notify_func)

    def test_create_with_config_none(self):
        path = os.path.join(self.test_dir, "cfg_repo")
        r = repos.create(path, config=None)
        self.assertIsNotNone(r)

    def test_create_with_fs_config_none(self):
        path = os.path.join(self.test_dir, "fscfg_repo")
        r = repos.create(path, fs_config=None)
        self.assertIsNotNone(r)

    def test_load_fs_with_parent_dir(self):
        r = repos.create(os.path.join(self.test_dir, "parentdir"))
        dumpfile = textwrap.dedent("""\
        SVN-fs-dump-format-version: 2

        UUID: 38f0a982-fd1f-4e00-aa6b-a20720f4b9ca

        Revision-number: 0
        Prop-content-length: 56
        Content-length: 56

        K 8
        svn:date
        V 27
        2011-08-26T13:08:30.187858Z
        PROPS-END
        """).encode("ascii")
        feedback = BytesIO()
        r.load_fs(BytesIO(dumpfile), feedback, repos.LOAD_UUID_DEFAULT, parent_dir="")

    def test_paths_changed(self):
        repos.create(os.path.join(self.test_dir, "foo"))
        root = repos.Repository("foo").fs().revision_root(0)
        self.assertEqual({}, root.paths_changed())

    def test_is_dir(self):
        repos.create(os.path.join(self.test_dir, "foo"))
        root = repos.Repository("foo").fs().revision_root(0)
        self.assertEqual(True, root.is_dir(""))
        # TODO(jelmer): Newer versions of libsvn_repos crash when passed a
        # nonexistant path.
        # self.assertEqual(False, root.is_dir("nonexistant"))

    def test_is_file(self):
        repos.create(os.path.join(self.test_dir, "foo"))
        root = repos.Repository("foo").fs().revision_root(0)
        self.assertEqual(False, root.is_file(""))
        # TODO(jelmer): Newer versions of libsvn_repos crash when passed a
        # nonexistant path.
        # self.assertEqual(False, root.is_file("nonexistant"))


class TestRepositoryDelete(TestCaseInTempDir):
    def test_delete(self):
        path = os.path.join(self.test_dir, "todelete")
        repos.create(path)
        self.assertTrue(os.path.exists(path))
        repos.delete(path)
        self.assertFalse(os.path.exists(path))


class TestRepositoryHotcopy(TestCaseInTempDir):
    def test_hotcopy(self):
        src = os.path.join(self.test_dir, "src")
        dest = os.path.join(self.test_dir, "dest")
        repos.create(src)
        repos.hotcopy(src, dest)
        self.assertTrue(os.path.exists(dest))
        # Verify the copy is a valid repository
        r = repos.Repository(dest)
        self.assertEqual(0, r.fs().youngest_revision())


class TestFileSystemRoot(TestCaseInTempDir):
    def setUp(self):
        super(TestFileSystemRoot, self).setUp()
        self.repo_path = os.path.join(self.test_dir, "repo")
        repos.create(self.repo_path)

    def test_file_length(self):
        # Commit a file first using the repos API
        r = repos.Repository(self.repo_path)
        root = r.fs().revision_root(0)
        # Root has no files, so we just test on what's available
        # after creation, the root dir exists
        self.assertTrue(root.is_dir(""))

    def test_proplist_root(self):
        r = repos.Repository(self.repo_path)
        root = r.fs().revision_root(0)
        props = root.proplist("")
        self.assertIsInstance(props, dict)


class TestFileSystemRootWithFile(SubversionTestCase):
    """Tests for FileSystemRoot methods that require a committed file."""

    def setUp(self):
        super(TestFileSystemRootWithFile, self).setUp()
        self.repos_url = self.make_repository("d")
        dc = self.get_commit_editor(self.repos_url)
        f = dc.add_file("testfile")
        f.modify(b"hello world")
        dc.close()

    def _get_repo_path(self):
        # file:// URL to path
        import sys

        if sys.platform == "win32":
            from urllib.request import url2pathname

            # On Windows, repos_url is file:///D:/... via pathname2url
            return url2pathname(self.repos_url[len("file:") :])
        return self.repos_url[len("file://") :]

    def test_file_length(self):
        r = repos.Repository(self._get_repo_path())
        root = r.fs().revision_root(1)
        length = root.file_length("testfile")
        self.assertEqual(11, length)

    def test_file_content(self):
        r = repos.Repository(self._get_repo_path())
        root = r.fs().revision_root(1)
        stream = root.file_content("testfile")
        data = stream.read()
        self.assertEqual(b"hello world", data)

    def test_file_checksum_md5(self):
        r = repos.Repository(self._get_repo_path())
        root = r.fs().revision_root(1)
        # kind=0 is svn_checksum_md5
        checksum = root.file_checksum("testfile", 0)
        self.assertIsInstance(checksum, str)
        self.assertEqual(32, len(checksum))

    def test_file_checksum_sha1(self):
        r = repos.Repository(self._get_repo_path())
        root = r.fs().revision_root(1)
        # kind=1 is svn_checksum_sha1
        checksum = root.file_checksum("testfile", 1)
        self.assertIsInstance(checksum, str)
        self.assertEqual(40, len(checksum))

    def test_file_checksum_force(self):
        r = repos.Repository(self._get_repo_path())
        root = r.fs().revision_root(1)
        checksum = root.file_checksum("testfile", 0, True)
        self.assertIsInstance(checksum, str)
        self.assertEqual(32, len(checksum))


class StreamTests(TestCase):
    def test_read(self):
        s = repos.Stream()
        self.assertEqual(b"", s.read())
        self.assertEqual(b"", s.read(15))
        s.close()

    def test_write(self):
        s = repos.Stream()
        self.assertEqual(0, s.write(b""))
        self.assertEqual(2, s.write(b"ab"))
        s.close()

    def test_close(self):
        s = repos.Stream()
        s.close()

    def test_read_full(self):
        s = repos.Stream()
        self.assertEqual(b"", s.read())
        s.close()


class ConstantsTests(TestCase):
    def test_load_uuid_constants(self):
        self.assertIsInstance(repos.LOAD_UUID_DEFAULT, int)
        self.assertIsInstance(repos.LOAD_UUID_IGNORE, int)
        self.assertIsInstance(repos.LOAD_UUID_FORCE, int)

    def test_path_change_constants(self):
        self.assertIsInstance(repos.PATH_CHANGE_MODIFY, int)
        self.assertIsInstance(repos.PATH_CHANGE_ADD, int)
        self.assertIsInstance(repos.PATH_CHANGE_DELETE, int)
        self.assertIsInstance(repos.PATH_CHANGE_REPLACE, int)

    def test_checksum_constants(self):
        self.assertIsInstance(repos.CHECKSUM_MD5, int)
        self.assertIsInstance(repos.CHECKSUM_SHA1, int)



# Copyright (C) 2005-2007 Jelmer Vernooij <jelmer@jelmer.uk>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2.1 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Handling of Subversion properties."""

__author__ = "Jelmer Vernooij <jelmer@jelmer.uk>"
__docformat__ = "restructuredText"


import bisect
import calendar
import time

try:
    import urlparse
except ImportError:
    import urllib.parse as urlparse


class InvalidExternalsDescription(Exception):
    _fmt = """Unable to parse externals description."""


def is_valid_property_name(prop):
    """Check the validity of a property name.

    :param prop: Property name
    :return: Whether prop is a valid property name
    """
    if not prop[0].isalnum() and prop[0] not in ":_":
        return False
    for c in prop[1:]:
        if not c.isalnum() and c not in "-:._":
            return False
    return True


def time_to_cstring(timestamp):
    """Determine string representation of a time.

    :param timestamp: Number of microseconds since the start of 1970
    :return: string with date
    """
    tm_usec = timestamp % 1000000
    (tm_year, tm_mon, tm_mday, tm_hour, tm_min, tm_sec, tm_wday, tm_yday, tm_isdst) = (
        time.gmtime(timestamp / 1000000)
    )
    return "%04d-%02d-%02dT%02d:%02d:%02d.%06dZ" % (
        tm_year,
        tm_mon,
        tm_mday,
        tm_hour,
        tm_min,
        tm_sec,
        tm_usec,
    )


def time_from_cstring(text):
    """Parse a time from a cstring.

    :param text: Parse text
    :return: number of microseconds since the start of 1970
    """
    (basestr, usecstr) = text.split(".", 1)
    assert usecstr[-1] == "Z"
    tm_usec = int(usecstr[:-1])
    tm = time.strptime(basestr, "%Y-%m-%dT%H:%M:%S")
    return int(calendar.timegm(tm)) * 1000000 + tm_usec


def parse_externals_description(base_url, val):
    """Parse an svn:externals property value.

    :param base_url: URL on which the property is set. Used for
        relative externals.

    :returns: dictionary with local names as keys, (revnum, url)
              as value. revnum is the revision number and is
              set to None if not applicable.
    """

    def is_url(u):
        return "://" in u

    ret = {}
    for line in val.splitlines():
        if line == "" or line[0] == "#":
            continue
        pts = line.rsplit(None, 3)
        if len(pts) == 4:
            if pts[0] == "-r":  # -r X URL DIR
                revno = int(pts[1])
                path = pts[3]
                relurl = pts[2]
            elif pts[1] == "-r":  # DIR -r X URL
                revno = int(pts[2])
                path = pts[0]
                relurl = pts[3]
            else:
                raise InvalidExternalsDescription()
        elif len(pts) == 3:
            if pts[1].startswith("-r"):  # DIR -rX URL
                revno = int(pts[1][2:])
                path = pts[0]
                relurl = pts[2]
            elif pts[0].startswith("-r"):  # -rX URL DIR
                revno = int(pts[0][2:])
                path = pts[2]
                relurl = pts[1]
            else:
                raise InvalidExternalsDescription()
        elif len(pts) == 2:
            if not is_url(pts[0]):
                relurl = pts[1]
                path = pts[0]
            else:
                relurl = pts[0]
                path = pts[1]
            revno = None
        else:
            raise InvalidExternalsDescription()
        if relurl.startswith("//"):
            raise NotImplementedError(
                "Relative to the scheme externals not yet supported"
            )
        if relurl.startswith("^/"):
            raise NotImplementedError(
                "Relative to the repository root externals not yet supported"
            )
        ret[path] = (revno, urlparse.urljoin(base_url + "/", relurl))
    return ret


def parse_mergeinfo_property(text):
    """Parse a mergeinfo property.

    :param text: Property contents
    """
    ret = {}
    for line in text.splitlines():
        (path, ranges) = line.rsplit(":", 1)
        assert path.startswith("/")
        ret[path] = []
        for range in ranges.split(","):
            if range[-1] == "*":
                inheritable = False
                range = range[:-1]
            else:
                inheritable = True
            try:
                (start, end) = range.split("-", 1)
                ret[path].append((int(start), int(end), inheritable))
            except ValueError:
                ret[path].append((int(range), int(range), inheritable))

    return ret


def generate_mergeinfo_property(merges):
    """Generate the contents of the svn:mergeinfo property

    :param merges: dictionary mapping paths to lists of ranges
    :return: Property contents
    """

    def formatrange(range_params):
        (start, end, inheritable) = range_params
        suffix = ""
        if not inheritable:
            suffix = "*"
        if start == end:
            return "%d%s" % (start, suffix)
        else:
            return "%d-%d%s" % (start, end, suffix)

    text = ""
    for path, ranges in merges.items():
        assert path.startswith("/")
        text += "%s:%s\n" % (path, ",".join(map(formatrange, ranges)))
    return text


def range_includes_revnum(ranges, revnum):
    """Check if the specified range contains the mentioned revision number.

    :param ranges: list of ranges
    :param revnum: revision number
    :return: Whether or not the revision number is included
    """
    i = bisect.bisect(ranges, (revnum, revnum, True))
    if i == 0:
        return False
    (start, end, inheritable) = ranges[i - 1]
    return start <= revnum <= end


def range_add_revnum(ranges, revnum, inheritable=True):
    """Add revision number to a list of ranges

    :param ranges: List of ranges
    :param revnum: Revision number to add
    :param inheritable: TODO
    :return: New list of ranges
    """
    # TODO: Deal with inheritable
    item = (revnum, revnum, inheritable)
    if len(ranges) == 0:
        ranges.append(item)
        return ranges
    i = bisect.bisect(ranges, item)
    if i > 0:
        (start, end, inh) = ranges[i - 1]
        if start <= revnum <= end:
            # already there
            return ranges
        if end == revnum - 1:
            # Extend previous range
            ranges[i - 1] = (start, end + 1, inh)
            return ranges
    if i < len(ranges):
        (start, end, inh) = ranges[i]
        if start - 1 == revnum:
            # Extend next range
            ranges[i] = (start - 1, end, inh)
            return ranges
    ranges.insert(i, item)
    return ranges


def mergeinfo_includes_revision(merges, path, revnum):
    """Check if the specified mergeinfo contains a path in revnum.

    :param merges: Dictionary with merges
    :param path: Merged path
    :param revnum: Revision number
    :return: Whether the revision is included
    """
    assert path.startswith("/")
    try:
        ranges = merges[path]
    except KeyError:
        return False

    return range_includes_revnum(ranges, revnum)


def mergeinfo_add_revision(mergeinfo, path, revnum):
    """Add a revision to a mergeinfo dictionary

    :param mergeinfo: Merginfo dictionary
    :param path: Merged path to add
    :param revnum: Merged revision to add
    :return: Updated dictionary
    """
    assert path.startswith("/")
    mergeinfo[path] = range_add_revnum(mergeinfo.get(path, []), revnum)
    return mergeinfo


PROP_EXECUTABLE = "svn:executable"
PROP_EXECUTABLE_VALUE = b"*"
PROP_EXTERNALS = "svn:externals"
PROP_IGNORE = "svn:ignore"
PROP_KEYWORDS = "svn:keywords"
PROP_MIME_TYPE = "svn:mime-type"
PROP_MERGEINFO = "svn:mergeinfo"
PROP_NEEDS_LOCK = "svn:needs-lock"
PROP_NEEDS_LOCK_VALUE = b"*"
PROP_PREFIX = "svn:"
PROP_SPECIAL = "svn:special"
PROP_SPECIAL_VALUE = b"*"
PROP_WC_PREFIX = "svn:wc:"
PROP_ENTRY_PREFIX = "svn:entry"
PROP_ENTRY_COMMITTED_DATE = "svn:entry:committed-date"
PROP_ENTRY_COMMITTED_REV = "svn:entry:committed-rev"
PROP_ENTRY_LAST_AUTHOR = "svn:entry:last-author"
PROP_ENTRY_LOCK_TOKEN = "svn:entry:lock-token"
PROP_ENTRY_UUID = "svn:entry:uuid"

PROP_REVISION_LOG = "svn:log"
PROP_REVISION_AUTHOR = "svn:author"
PROP_REVISION_DATE = "svn:date"
PROP_REVISION_ORIGINAL_DATE = "svn:original-date"


def diff(current, previous):
    """Find the differences between two property dictionaries.

    :param current: Dictionary with current (new) properties
    :param previous: Dictionary with previous (old) properties
    :return: Dictionary that contains an entry for
             each property that was changed. Value is a tuple
             with the old and the new property value.
    """
    ret = {}
    for key, newval in current.items():
        oldval = previous.get(key)
        if oldval != newval:
            ret[key] = (oldval, newval)
    return ret



# Copyright (C) 2006-2017 Jelmer Vernooij <jelmer@jelmer.uk>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2.1 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA

"""Python bindings for Subversion."""

import os
import platform

__author__ = "Jelmer Vernooij <jelmer@jelmer.uk>"
__version__ = (0, 11, 1)

NODE_DIR = 2
NODE_FILE = 1
NODE_NONE = 0
NODE_UNKNOWN = 3

ERR_UNSUPPORTED_FEATURE = 200007
ERR_RA_SVN_UNKNOWN_CMD = 210001
ERR_RA_SVN_CONNECTION_CLOSED = 210002
ERR_WC_LOCKED = 155004
ERR_RA_NOT_AUTHORIZED = 170001
ERR_INCOMPLETE_DATA = 200003
ERR_DIR_NOT_EMPTY = 200011
ERR_RA_SVN_MALFORMED_DATA = 210004
ERR_RA_NOT_IMPLEMENTED = 170003
ERR_FS_NO_SUCH_REVISION = 160006
ERR_FS_TXN_OUT_OF_DATE = 160028
ERR_REPOS_DISABLED_FEATURE = 165006
ERR_STREAM_MALFORMED_DATA = 140001
ERR_RA_ILLEGAL_URL = 170000
ERR_RA_LOCAL_REPOS_OPEN_FAILED = 180001
ERR_BAD_FILENAME = 125001
ERR_BAD_URL = 125002
ERR_BAD_DATE = 125003
ERR_RA_DAV_REQUEST_FAILED = 175002
ERR_RA_DAV_PATH_NOT_FOUND = 175007
ERR_FS_NOT_DIRECTORY = 160016
ERR_FS_NOT_FOUND = 160013
ERR_FS_ALREADY_EXISTS = 160020
ERR_RA_SVN_REPOS_NOT_FOUND = 210005
ERR_WC_NOT_WORKING_COPY = ERR_WC_NOT_DIRECTORY = 155007
ERR_ENTRY_EXISTS = 150002
ERR_WC_PATH_NOT_FOUND = 155010
ERR_CANCELLED = 200015
ERR_WC_UNSUPPORTED_FORMAT = 155021
ERR_UNKNOWN_CAPABILITY = 200026
ERR_AUTHN_NO_PROVIDER = 215001
ERR_RA_DAV_RELOCATED = 175011
ERR_FS_NOT_FILE = 160017
ERR_WC_BAD_ADM_LOG = 155009
ERR_WC_BAD_ADM_LOG_START = 155020
ERR_WC_NOT_LOCKED = 155005
ERR_RA_DAV_NOT_VCC = 20014
ERR_REPOS_HOOK_FAILURE = 165001
ERR_XML_MALFORMED = 130003
ERR_MALFORMED_FILE = 200002
ERR_FS_PATH_SYNTAX = 160005
ERR_RA_DAV_FORBIDDEN = 175013
ERR_WC_SCHEDULE_CONFLICT = 155013
ERR_RA_DAV_PROPPATCH_FAILED = 175008
ERR_SVNDIFF_CORRUPT_WINDOW = 185001
ERR_FS_CONFLICT = 160024
ERR_NODE_UNKNOWN_KIND = 145000
ERR_RA_SERF_SSL_CERT_UNTRUSTED = 230001
ERR_ENTRY_NOT_FOUND = 150000
ERR_BAD_PROPERTY_VALUE = 125005
ERR_FS_ROOT_DIR = 160021
ERR_WC_NODE_KIND_CHANGE = 155018
ERR_WC_UPGRADE_REQUIRED = 155036
ERR_RA_CANNOT_CREATE_SESSION = 170013
ERR_REPOS_BAD_ARGS = 165002

ERR_APR_OS_START_EAIERR = 670000
ERR_APR_OS_ERRSPACE_SIZE = 50000
ERR_CATEGORY_SIZE = 5000


# These will be removed in the next version of subvertpy
ERR_EAI_NONAME = 670008
ERR_UNKNOWN_HOSTNAME = 670002

AUTH_PARAM_DEFAULT_USERNAME = "svn:auth:username"
AUTH_PARAM_DEFAULT_PASSWORD = "svn:auth:password"

SSL_NOTYETVALID = 0x00000001
SSL_EXPIRED = 0x00000002
SSL_CNMISMATCH = 0x00000004
SSL_UNKNOWNCA = 0x00000008
SSL_OTHER = 0x40000000


class SubversionException(Exception):
    """A Subversion exception"""

    def __init__(self, msg, num, child=None, location=None):
        self.args = (msg, num)
        self.child = child
        self.location = location


def _check_mtime(m):
    """Check whether a C extension is out of date.

    :param m: Python module that is a C extension
    """
    import os

    (base, _) = os.path.splitext(m.__file__)
    c_file = "%s.c" % base
    if not os.path.exists(c_file):
        return True
    if os.path.getmtime(m.__file__) < os.path.getmtime(c_file):
        return False
    return True


if platform.system() == "Windows":
    dll_dirs = [
        dll_dir
        for dll_dir in os.environ.get("SUBVERTPY_DLL_PATH", "").split(";")
        if dll_dir
    ]
    for path in dll_dirs:
        path = os.path.abspath(path)
        if os.path.exists(path):
            os.add_dll_directory(path)

try:
    from subvertpy import client, _ra, repos, wc

    for x in client, _ra, repos, wc:
        if not _check_mtime(x):
            from warnings import warn

            warn("subvertpy extensions are outdated and need to be rebuilt")
            break
except ImportError as e:
    raise ImportError("Unable to load subvertpy extensions: %s" % e)



# Copyright (C) 2006-2007 Jelmer Vernooij <jelmer@jelmer.uk>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA

"""Tests for subvertpy.marshall."""

from subvertpy.marshall import (
    MarshallError,
    NeedMoreData,
    literal,
    marshall,
    unmarshall,
)
from subvertpy.tests import TestCase


class TestMarshalling(TestCase):
    def test_literal_txt(self):
        line = literal("foo")
        self.assertEqual("foo", line.txt)

    def test_literal_str(self):
        line = literal("foo bar")
        self.assertEqual("foo bar", line.__str__())

    def test_literal_rep(self):
        line = literal("foo bar")
        self.assertEqual("foo bar", line.__repr__())

    def test_literal_eq_same(self):
        self.assertEqual(literal("foo"), literal("foo"))

    def test_literal_eq_different(self):
        self.assertNotEqual(literal("foo"), literal("bar"))

    def test_literal_eq_non_literal(self):
        self.assertNotEqual(literal("foo"), "foo")

    def test_marshall_error(self):
        err = MarshallError("bla bla")
        self.assertEqual("bla bla", err.__str__())

    def test_need_more_data_is_marshall_error(self):
        self.assertTrue(issubclass(NeedMoreData, MarshallError))

    def test_need_more_data(self):
        err = NeedMoreData("need more")
        self.assertIsInstance(err, MarshallError)
        self.assertEqual("need more", str(err))

    def test_marshall_int(self):
        self.assertEqual(b"1 ", marshall(1))

    def test_marshall_int_zero(self):
        self.assertEqual(b"0 ", marshall(0))

    def test_marshall_int_large(self):
        self.assertEqual(b"999999 ", marshall(999999))

    def test_marshall_list(self):
        self.assertEqual(b"( 1 2 3 4 ) ", marshall([1, 2, 3, 4]))

    def test_marshall_list_mixed(self):
        self.assertEqual(b"( 1 3 4 3:str ) ", marshall([1, 3, 4, "str"]))

    def test_marshall_literal(self):
        self.assertEqual(b"foo ", marshall(literal("foo")))

    def test_marshall_string(self):
        self.assertEqual(b"3:foo ", marshall("foo"))

    def test_marshall_bytes(self):
        self.assertEqual(b"3:foo ", marshall(b"foo"))

    def test_marshall_bytes_binary(self):
        data = b"\x00\x01\x02"
        self.assertEqual(b"3:\x00\x01\x02 ", marshall(data))

    def test_marshall_string_utf8(self):
        # UTF-8 multi-byte characters
        result = marshall("\u00e9")  # é
        self.assertEqual(b"2:\xc3\xa9 ", result)

    def test_marshall_raises(self):
        self.assertRaises(MarshallError, marshall, dict())

    def test_marshall_list_nested(self):
        self.assertEqual(b"( ( ( 3 ) 4 ) ) ", marshall([[[3], 4]]))

    def test_marshall_list_empty(self):
        self.assertEqual(b"( ) ", marshall([]))

    def test_marshall_tuple(self):
        self.assertEqual(b"( 1 2 ) ", marshall((1, 2)))

    def test_marshall_string_space(self):
        self.assertEqual(b"5:bla l ", marshall("bla l"))

    def test_marshall_string_empty(self):
        self.assertEqual(b"0: ", marshall(""))

    def test_unmarshall_string(self):
        self.assertEqual((b"", b"bla l"), unmarshall(b"5:bla l"))

    def test_unmarshall_list(self):
        self.assertEqual((b"", [4, 5]), unmarshall(b"( 4 5 ) "))

    def test_unmarshall_list_empty(self):
        self.assertEqual((b"", []), unmarshall(b"( ) "))

    def test_unmarshall_list_nested(self):
        self.assertEqual((b"", [[1, 2], 3]), unmarshall(b"( ( 1 2 ) 3 ) "))

    def test_unmarshall_int(self):
        self.assertEqual((b"", 2), unmarshall(b"2 "))

    def test_unmarshall_int_zero(self):
        self.assertEqual((b"", 0), unmarshall(b"0 "))

    def test_unmarshall_int_large(self):
        self.assertEqual((b"", 123456), unmarshall(b"123456 "))

    def test_unmarshall_literal(self):
        self.assertEqual((b"", literal("x")), unmarshall(b"x "))

    def test_unmarshall_literal_with_digits(self):
        self.assertEqual((b"", literal("foo2")), unmarshall(b"foo2 "))

    def test_unmarshall_literal_with_dash(self):
        self.assertEqual((b"", literal("foo-bar")), unmarshall(b"foo-bar "))

    def test_unmarshall_empty(self):
        self.assertRaises(NeedMoreData, unmarshall, b"")

    def test_unmarshall_nospace(self):
        self.assertRaises(MarshallError, unmarshall, b"nospace")

    def test_unmarshall_toolong(self):
        self.assertRaises(NeedMoreData, unmarshall, b"43432432:bla")

    def test_unmarshall_literal_negative(self):
        self.assertRaises(MarshallError, unmarshall, b":-3213")

    def test_unmarshall_open_list(self):
        self.assertRaises(NeedMoreData, unmarshall, b"( 3 4 ")

    def test_unmarshall_remaining_data(self):
        # Two marshalled items concatenated
        data = marshall(b"foo") + marshall(b"bar")
        remainder, value = unmarshall(data)
        self.assertEqual(b"foo", value)
        _, value2 = unmarshall(remainder)
        self.assertEqual(b"bar", value2)

    def test_unmarshall_list_missing_space_after_open(self):
        self.assertRaises(MarshallError, unmarshall, b"(x")

    def test_unmarshall_need_more_data_list_start(self):
        self.assertRaises(NeedMoreData, unmarshall, b"(")

    def test_roundtrip_int(self):
        _, val = unmarshall(marshall(42))
        self.assertEqual(42, val)

    def test_roundtrip_string(self):
        _, val = unmarshall(marshall(b"hello"))
        self.assertEqual(b"hello", val)

    def test_roundtrip_list(self):
        _, val = unmarshall(marshall([1, 2, 3]))
        self.assertEqual([1, 2, 3], val)

    def test_roundtrip_literal(self):
        _, val = unmarshall(marshall(literal("success")))
        self.assertEqual(literal("success"), val)

    def test_roundtrip_nested(self):
        _, val = unmarshall(marshall([literal("cmd"), [1, b"data"]]))
        self.assertEqual([literal("cmd"), [1, b"data"]], val)

    def test_marshall_bool_true(self):
        # In Python 3, bool is a subclass of int, so True marshalls as 1
        self.assertEqual(b"1 ", marshall(True))

    def test_marshall_bool_false(self):
        # In Python 3, bool is a subclass of int, so False marshalls as 0
        self.assertEqual(b"0 ", marshall(False))

    def test_marshall_none_raises(self):
        self.assertRaises(MarshallError, marshall, None)

    def test_marshall_float_raises(self):
        self.assertRaises(MarshallError, marshall, 3.14)

    def test_unmarshall_literal_newline_separator(self):
        # newline is also valid whitespace
        self.assertEqual((b"", literal("x")), unmarshall(b"x\n"))

    def test_unmarshall_int_newline_separator(self):
        self.assertEqual((b"", 5), unmarshall(b"5\n"))

    def test_unmarshall_list_newline_separator(self):
        self.assertEqual((b"", [1]), unmarshall(b"( 1 )\n"))



# Copyright (C) 2006-2007 Jelmer Vernooij <jelmer@jelmer.uk>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA

"""Marshalling for the svn_ra protocol."""


class literal(object):
    """A protocol literal."""

    def __init__(self, txt):
        self.txt = txt

    def __str__(self):
        return self.txt

    def __repr__(self):
        return self.txt

    def __eq__(self, other):
        return isinstance(other, literal) and self.txt == other.txt


# 1. Syntactic structure
# ----------------------
#
# The Subversion protocol is specified in terms of the following
# syntactic elements, specified using ABNF [RFC 2234]:
#
#   item   = word / number / string / list
#   word   = ALPHA *(ALPHA / DIGIT / "-") space
#   number = 1*DIGIT space
#   string = 1*DIGIT ":" *OCTET space
#          ; digits give the byte count of the *OCTET portion
#   list   = "(" space *item ")" space
#   space  = 1*(SP / LF)
#


class MarshallError(Exception):
    """A Marshall error."""


class NeedMoreData(MarshallError):
    """More data needed."""


def marshall(x):
    """Marshall a Python data item.

    :param x: Data item
    :return: encoded byte string
    """
    if isinstance(x, int):
        return ("%d " % x).encode("ascii")
    elif isinstance(x, (list, tuple)):
        return b"( " + bytes().join(map(marshall, x)) + b") "
    elif isinstance(x, literal):
        return ("%s " % x).encode("ascii")
    elif isinstance(x, bytes):
        return ("%d:" % len(x)).encode("ascii") + x + b" "
    elif isinstance(x, str):
        x = x.encode("utf-8")
        return ("%d:" % len(x)).encode("ascii") + x + b" "
    elif isinstance(x, bool):
        if x is True:
            return b"true "
        elif x is False:
            return b"false "
    raise MarshallError("Unable to marshall type %s" % x)


def unmarshall(x):
    """Unmarshall the next item from a buffer.

    :param x: Bytes to parse
    :return: tuple with unpacked item and remaining bytes
    """
    whitespace = frozenset(b"\n ")
    if len(x) == 0:
        raise NeedMoreData("Not enough data")
    if x[0:1] == b"(":  # list follows
        if len(x) <= 1:
            raise NeedMoreData("Missing whitespace")
        if x[1:2] != b" ":
            raise MarshallError("missing whitespace after list start")
        x = x[2:]
        ret = []
        try:
            while x[0:1] != b")":
                (x, n) = unmarshall(x)
                ret.append(n)
        except IndexError:
            raise NeedMoreData("List not terminated")

        if len(x) <= 1:
            raise NeedMoreData("Missing whitespace")

        if x[1] not in whitespace:
            raise MarshallError("Expected space, got '%c'" % x[1])

        return (x[2:], ret)
    elif x[0:1].isdigit():
        num = bytearray()
        # Check if this is a string or a number
        while x[:1].isdigit():
            num.append(x[0])
            x = x[1:]
        num = int(num)

        if x[0] in whitespace:
            return (x[1:], num)
        elif x[0:1] == b":":
            if len(x) < num:
                raise NeedMoreData("Expected string of length %r" % num)
            return (x[num + 2 :], x[1 : num + 1])
        elif not x:
            raise MarshallError("Expected whitespace, got end of string.")
        else:
            raise MarshallError("Expected whitespace or ':', got '%c'" % x[0])
    elif x[:1].isalpha():
        ret = bytearray()
        # Parse literal
        try:
            while x[:1].isalpha() or x[:1].isdigit() or x[0:1] == b"-":
                ret.append(x[0])
                x = x[1:]
        except IndexError:
            raise NeedMoreData("Expected literal")

        if not x:
            raise MarshallError("Expected whitespace, got end of string.")

        if x[0] not in whitespace:
            raise MarshallError("Expected whitespace, got '%c'" % x[0])

        return (x[1:], literal(ret.decode("ascii")))
    else:
        raise MarshallError("Unexpected character '%c'" % x[0])



#!/usr/bin/env python
# Setup file for subvertpy
# Copyright (C) 2005-2010 Jelmer Vernooij <jelmer@jelmer.uk>

from setuptools import setup
from setuptools.extension import Extension
import errno
import sys
import os
import re
import shlex
import subprocess


def split_shell_results(line):
    return shlex.split(line)


def config_value(command, env, args):
    command = os.environ.get(env, command)
    try:
        return subprocess.check_output([command] + args).strip().decode()
    except OSError as e:
        if e.errno == errno.ENOENT:
            raise Exception(
                "%s not found. Please set %s environment variable" % (command, env)
            )
        raise


def apr_config(args):
    return config_value("apr-1-config", "APR_CONFIG", args)


def apu_config(args):
    return config_value("apu-1-config", "APU_CONFIG", args)


def apr_build_data():
    """Determine the APR header file location."""
    try:
        includedir = os.environ["APR_INCLUDE_DIR"]
    except KeyError:
        if os.name != "nt":
            includedir = apr_config(["--includedir"])
    if not os.path.isdir(includedir):
        raise Exception("APR development headers not found")
    try:
        extra_link_flags = split_shell_results(os.environ["APR_LINK_FLAGS"])
    except KeyError:
        if os.name != "nt":
            extra_link_flags = split_shell_results(apr_config(["--link-ld", "--libs"]))
        else:
            extra_link_flags = []
    return (includedir, extra_link_flags)


def apu_build_data():
    """Determine the APR util header file location."""
    try:
        includedir = os.environ["APU_INCLUDE_DIR"]
    except KeyError:
        if os.name != "nt":
            includedir = apu_config(["--includedir"])
    if not os.path.isdir(includedir):
        raise Exception("APR util development headers not found")
    try:
        extra_link_flags = split_shell_results(os.environ["APU_LINK_FLAGS"])
    except KeyError:
        if os.name != "nt":
            extra_link_flags = split_shell_results(apu_config(["--link-ld", "--libs"]))
        else:
            extra_link_flags = []
    return (includedir, extra_link_flags)


def svn_build_data():
    """Determine the Subversion header file location."""
    if "SVN_HEADER_PATH" in os.environ and "SVN_LIBRARY_PATH" in os.environ:
        return ([os.getenv("SVN_HEADER_PATH")], [os.getenv("SVN_LIBRARY_PATH")], [])
    svn_prefix = os.getenv("SVN_PREFIX")
    if svn_prefix is None:
        basedirs = ["/usr/local", "/usr", "/opt/homebrew", "/opt/local"]
        for basedir in basedirs:
            includedir = os.path.join(basedir, "include/subversion-1")
            if os.path.isdir(includedir):
                svn_prefix = basedir
                break
    if svn_prefix is not None:
        return (
            [os.path.join(svn_prefix, "include/subversion-1")],
            [os.path.join(svn_prefix, "lib"), os.path.join(svn_prefix, "lib", "db48")],
            [],
        )
    raise Exception(
        "Subversion development files not found. "
        "Please set SVN_PREFIX or (SVN_LIBRARY_PATH and "
        "SVN_HEADER_PATH) environment variable. "
    )


def is_keychain_provider_available():
    """
    Checks for the availability of the Keychain simple authentication provider
    in Subversion by compiling a simple test program.
    """
    abd = apr_build_data()
    sbd = svn_build_data()
    gcc_command_args = (
        ["gcc"]
        + ["-I" + inc for inc in sbd[0]]
        + ["-L" + lib for lib in sbd[1]]
        + ["-I" + abd[0], "-lsvn_subr-1", "-x", "c", "-"]
    )
    gcc = subprocess.Popen(
        gcc_command_args,
        stdin=subprocess.PIPE,
        universal_newlines=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    gcc.communicate("""
#include <svn_auth.h>
int main(int argc, const char* arv[]) {
    svn_auth_get_keychain_simple_provider(NULL, NULL);
}
""")
    return gcc.returncode == 0


class VersionQuery(object):
    def __init__(self, filename):
        self.filename = filename
        f = open(filename, "rU")
        try:
            self.text = f.read()
        finally:
            f.close()

    def grep(self, what):
        m = re.search(r"^#define\s+%s\s+(\d+)\s*$" % (what,), self.text, re.MULTILINE)
        if not m:
            raise Exception(
                "Definition for %s was not found in file %s." % (what, self.filename)
            )
        return int(m.group(1))


(apr_includedir, apr_link_flags) = apr_build_data()
(apu_includedir, apu_link_flags) = apu_build_data()
(svn_includedirs, svn_libdirs, svn_link_flags) = svn_build_data()


class SvnExtension(Extension):
    def __init__(self, name, *args, **kwargs):
        if sys.platform == "win32":
            libraries = kwargs.get("libraries", [])
            libs = [("lib" + lib) for lib in libraries]
            modified = True
            while modified:
                modified = False
                for lib in libraries:
                    for extra in deep_deps.get(lib, []):
                        if extra not in libs:
                            modified = True
                            libs.append(extra)
            kwargs["libraries"] = libs
        kwargs["include_dirs"] = (
            [apr_includedir, apu_includedir] + svn_includedirs + ["subvertpy"]
        )
        kwargs["library_dirs"] = svn_libdirs
        # Note that the apr-util link flags are not included here, as
        # subvertpy only uses some apr util constants but does not use
        # the library directly.
        kwargs["extra_link_args"] = apr_link_flags + apu_link_flags + svn_link_flags
        if os.name == "nt":
            # APR needs WIN32 defined.
            kwargs["define_macros"] = [("WIN32", None)]
        if sys.platform == "darwin":
            # on Mac OS X, we need to check for Keychain availability
            if is_keychain_provider_available():
                if "define_macros" not in kwargs:
                    kwargs["define_macros"] = []
                kwargs["define_macros"].extend(
                    (("DARWIN", None), ("SVN_KEYCHAIN_PROVIDER_AVAILABLE", "1"))
                )
        if sys.platform in ("darwin", "linux"):
            kwargs["extra_compile_args"] = [
                "-std=c99"  # for GCC >=15 as it is using c23 by default
            ]
        Extension.__init__(self, name, *args, **kwargs)


def source_path(filename):
    return os.path.join("subvertpy", filename)


# Urgh. It's a pain having to maintain these manually. But what else can we do?
subr_deep_deps = [
    "libsvn_subr-1",
    "libapr-1",
    "libaprutil-1",
    "sqlite3",
    "zlib",
    "advapi32",
    "crypt32",
    "libexpat",
    "shell32",
    "ws2_32",
    "mswsock",
    "version",
    "ole32",
]


repos_deep_deps = [
    "libsvn_repos-1",
    "libsvn_fs-1",
    "libsvn_fs_util-1",
    "libsvn_fs_fs-1",
    "libsvn_fs_x-1",
    "libsvn_delta-1",
    "libapr-1",
    "libaprutil-1",
]


ra_deep_deps = [
    "libsvn_ra_svn-1",
    "libsvn_ra_local-1",
    "libsvn_repos-1",
    "libapr-1",
    "libaprutil-1",
]


deep_deps = {
    "svn_ra-1": ra_deep_deps,
    "svn_repos-1": repos_deep_deps,
    "svn_subr-1": subr_deep_deps,
}


def subvertpy_modules():
    return [
        SvnExtension(
            "subvertpy.client",
            [
                source_path(n)
                for n in ("client.c", "editor.c", "util.c", "_ra.c", "wc.c")
            ],
            libraries=[
                "svn_client-1",
                "svn_diff-1",
                "svn_delta-1",
                "svn_wc-1",
                "svn_ra-1",
                "svn_subr-1",
            ],
        ),
        SvnExtension(
            "subvertpy._ra",
            [source_path(n) for n in ("_ra.c", "util.c", "editor.c")],
            libraries=["svn_delta-1", "svn_ra-1", "svn_subr-1"],
        ),
        SvnExtension(
            "subvertpy.repos",
            [source_path(n) for n in ("repos.c", "util.c")],
            libraries=["svn_repos-1", "svn_subr-1"],
        ),
        SvnExtension(
            "subvertpy.wc",
            [source_path(n) for n in ["wc.c", "util.c", "editor.c"]],
            libraries=["svn_wc-1", "svn_diff-1", "svn_delta-1", "svn_subr-1"],
        ),
        SvnExtension(
            "subvertpy.subr",
            [source_path(n) for n in ["util.c", "subr.c"]],
            libraries=["svn_subr-1"],
        ),
    ]


def package_data():
    if sys.platform == "win32":
        return {"subvertpy": ["subvertpy/*.dll"]}
    else:
        return {}


subvertpy_version = (0, 11, 1)
subvertpy_version_string = ".".join(map(str, subvertpy_version))


if __name__ == "__main__":
    setup(
        name="subvertpy",
        description="Alternative Python bindings for Subversion",
        keywords="svn subvertpy subversion bindings",
        version=subvertpy_version_string,
        url="https://jelmer.uk/subvertpy",
        download_url="https://jelmer.uk/subvertpy/tarball/subvertpy-%s/"
        % (subvertpy_version_string,),
        license="LGPLv2.1 or later",
        author="Jelmer Vernooij",
        author_email="jelmer@jelmer.uk",
        long_description="""
Alternative Python bindings for Subversion. The goal is to have
complete, portable and "Pythonic" Python bindings.

Bindings are provided for the working copy, client, delta, remote access and
repository APIs. A hookable server side implementation of the custom Subversion
protocol (svn_ra) is also provided.

Differences with similar packages
---------------------------------
subvertpy covers more of the APIs than python-svn. It provides a more
"Pythonic" API than python-subversion, which wraps the Subversion C API pretty
much directly. Neither provide a hookable server-side.

Dependencies
------------
Subvertpy depends on Python 2.7 or 3.5, and Subversion 1.4 or later. It should
work on Windows as well as most POSIX-based platforms (including Linux, BSDs
and Mac OS X).
""",
        packages=["subvertpy", "subvertpy.tests"],
        package_data=package_data(),
        ext_modules=subvertpy_modules(),
        scripts=["bin/subvertpy-fast-export"],
        classifiers=[
            "Development Status :: 4 - Beta",
            "License :: OSI Approved :: GNU General Public "
            "License v2 or later (GPLv2+)",
            "Programming Language :: Python :: 2.7",
            "Programming Language :: Python :: 3.4",
            "Programming Language :: Python :: 3.5",
            "Programming Language :: Python :: 3.6",
            "Programming Language :: Python :: Implementation :: CPython",
            "Operating System :: POSIX",
            "Topic :: Software Development :: Version Control",
        ],
    )



# Copyright (C) 2006 Jelmer Vernooij <jelmer@jelmer.uk>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA

"""Server backend base classes."""


class ServerBackend(object):
    """A server backend."""

    def open_repository(self, location):
        raise NotImplementedError(self.open_repository)


def generate_random_id():
    """Create a UUID for a repository."""
    import uuid

    return str(uuid.uuid4())


class ServerRepositoryBackend(object):
    def get_uuid(self):
        raise NotImplementedError(self.get_uuid)

    def get_latest_revnum(self):
        raise NotImplementedError(self.get_latest_revnum)

    def log(
        self,
        send_revision,
        target_path,
        start_rev,
        end_rev,
        changed_paths,
        strict_node,
        limit,
    ):
        raise NotImplementedError(self.log)

    def update(self, editor, revnum, target_path, recurse=True):
        raise NotImplementedError(self.update)

    def check_path(self, path, revnum):
        raise NotImplementedError(self.check_path)

    def stat(self, path, revnum):
        """Stat a path.

        Should return a dictionary with the following keys: name, kind, size,
            has-props, created-rev, created-date, last-author.
        """
        raise NotImplementedError(self.stat)

    def rev_proplist(self, revnum):
        raise NotImplementedError(self.rev_proplist)

    def get_locations(self, path, peg_revnum, revnums):
        raise NotImplementedError(self.get_locations)



# Copyright (C) 2005-2007 Jelmer Vernooij <jelmer@jelmer.uk>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2.1 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Subversion core library tests."""

import os
import time
from unittest import SkipTest
from subvertpy import properties
from subvertpy.tests import (
    TestCase,
)


class TestProperties(TestCase):
    def setUp(self):
        super(TestProperties, self).setUp()

    def test_time_from_cstring(self):
        self.assertEqual(
            1225704780716938,
            properties.time_from_cstring("2008-11-03T09:33:00.716938Z"),
        )

    def test_time_from_cstring_independent_from_dst(self):
        old_tz = os.environ.get("TZ", None)
        # On Windows, there is no tzset function, so skip this test.
        if getattr(time, "tzset", None) is None:
            raise SkipTest("tzset not available on Windows")

        try:
            # First specify a fixed timezone with known DST (late March to late
            # October)
            os.environ["TZ"] = "Europe/London"
            time.tzset()
            # Now test a time within that DST
            self.assertEqual(
                1275295762430000,
                properties.time_from_cstring("2010-05-31T08:49:22.430000Z"),
            )
        finally:
            if old_tz is None:
                del os.environ["TZ"]
            else:
                os.environ["TZ"] = old_tz
            time.tzset()

    def test_time_to_cstring(self):
        self.assertEqual(
            "2008-11-03T09:33:00.716938Z", properties.time_to_cstring(1225704780716938)
        )


class TestExternalsParser(TestCase):
    def test_parse_root_relative_externals(self):
        self.assertRaises(
            NotImplementedError,
            properties.parse_externals_description,
            "http://example.com",
            "third-party/skins              ^/foo",
        )

    def test_parse_scheme_relative_externals(self):
        self.assertRaises(
            NotImplementedError,
            properties.parse_externals_description,
            "http://example.com",
            "third-party/skins              //foo",
        )

    def test_parse_externals(self):
        self.assertEqual(
            {
                "third-party/sounds": (None, "http://sounds.red-bean.com/repos"),
                "third-party/skins": (
                    None,
                    "http://skins.red-bean.com/repositories/skinproj",
                ),
                "third-party/skins/toolkit": (
                    21,
                    "http://svn.red-bean.com/repos/skin-maker",
                ),
            },
            properties.parse_externals_description(
                "http://example.com",
                """\
third-party/sounds            http://sounds.red-bean.com/repos
third-party/skins          http://skins.red-bean.com/repositories/skinproj
third-party/skins/toolkit -r21 http://svn.red-bean.com/repos/skin-maker""",
            ),
        )

    def test_parse_externals_space_revno(self):
        self.assertEqual(
            {
                "third-party/skins/toolkit": (
                    21,
                    "http://svn.red-bean.com/repos/skin-maker",
                )
            },
            properties.parse_externals_description(
                "http://example.com",
                """\
third-party/skins/toolkit -r 21 http://svn.red-bean.com/repos/skin-maker""",
            ),
        )

    def test_parse_externals_swapped(self):
        self.assertEqual(
            {"third-party/sounds": (None, "http://sounds.red-bean.com/repos")},
            properties.parse_externals_description(
                "http://example.com",
                """\
http://sounds.red-bean.com/repos         third-party/sounds
""",
            ),
        )

    def test_parse_comment(self):
        self.assertEqual(
            {"third-party/sounds": (None, "http://sounds.red-bean.com/repos")},
            properties.parse_externals_description(
                "http://example.com/",
                """\

third-party/sounds             http://sounds.red-bean.com/repos
#third-party/skins              http://skins.red-bean.com/repositories/skinproj
#third-party/skins/toolkit -r21 http://svn.red-bean.com/repos/skin-maker""",
            ),
        )

    def test_parse_relative(self):
        self.assertEqual(
            {
                "third-party/sounds": (None, "http://example.com/branches/other"),
            },
            properties.parse_externals_description(
                "http://example.com/trunk",
                "third-party/sounds             ../branches/other",
            ),
        )

    def test_parse_repos_root_relative(self):
        self.assertEqual(
            {
                "third-party/sounds": (
                    None,
                    "http://example.com/bar/bla/branches/other",
                ),
            },
            properties.parse_externals_description(
                "http://example.com/trunk",
                "third-party/sounds             /bar/bla/branches/other",
            ),
        )

    def test_parse_invalid_missing_url(self):
        """No URL specified."""
        self.assertRaises(
            properties.InvalidExternalsDescription,
            lambda: properties.parse_externals_description(
                "http://example.com/", "bla"
            ),
        )

    def test_parse_invalid_too_much_data(self):
        """No URL specified."""
        self.assertRaises(
            properties.InvalidExternalsDescription,
            lambda: properties.parse_externals_description(
                None, "bla -R40 http://bla/"
            ),
        )


class MergeInfoPropertyParserTests(TestCase):
    def test_simple_range(self):
        self.assertEqual(
            {"/trunk": [(1, 2, True)]},
            properties.parse_mergeinfo_property("/trunk:1-2\n"),
        )

    def test_simple_range_uninheritable(self):
        self.assertEqual(
            {"/trunk": [(1, 2, False)]},
            properties.parse_mergeinfo_property("/trunk:1-2*\n"),
        )

    def test_simple_individual(self):
        self.assertEqual(
            {"/trunk": [(1, 1, True)]},
            properties.parse_mergeinfo_property("/trunk:1\n"),
        )

    def test_empty(self):
        self.assertEqual({}, properties.parse_mergeinfo_property(""))


class MergeInfoPropertyCreatorTests(TestCase):
    def test_simple_range(self):
        self.assertEqual(
            "/trunk:1-2\n",
            properties.generate_mergeinfo_property({"/trunk": [(1, 2, True)]}),
        )

    def test_simple_individual(self):
        self.assertEqual(
            "/trunk:1\n",
            properties.generate_mergeinfo_property({"/trunk": [(1, 1, True)]}),
        )

    def test_empty(self):
        self.assertEqual("", properties.generate_mergeinfo_property({}))


class RevnumRangeTests(TestCase):
    def test_add_revnum_empty(self):
        self.assertEqual([(1, 1, True)], properties.range_add_revnum([], 1))

    def test_add_revnum_before(self):
        self.assertEqual(
            [(2, 2, True), (8, 8, True)], properties.range_add_revnum([(2, 2, True)], 8)
        )

    def test_add_revnum_included(self):
        self.assertEqual([(1, 3, True)], properties.range_add_revnum([(1, 3, True)], 2))

    def test_add_revnum_after(self):
        self.assertEqual(
            [(1, 3, True), (5, 5, True)], properties.range_add_revnum([(1, 3, True)], 5)
        )

    def test_add_revnum_extend_before(self):
        self.assertEqual([(1, 3, True)], properties.range_add_revnum([(2, 3, True)], 1))

    def test_add_revnum_extend_after(self):
        self.assertEqual([(1, 3, True)], properties.range_add_revnum([(1, 2, True)], 3))

    def test_revnum_includes_empty(self):
        self.assertFalse(properties.range_includes_revnum([], 2))

    def test_revnum_includes_oor(self):
        self.assertFalse(
            properties.range_includes_revnum([(1, 3, True), (4, 5, True)], 10)
        )

    def test_revnum_includes_in(self):
        self.assertTrue(
            properties.range_includes_revnum([(1, 3, True), (4, 5, True)], 2)
        )


class MergeInfoIncludeTests(TestCase):
    def test_includes_individual(self):
        self.assertTrue(
            properties.mergeinfo_includes_revision(
                {"/trunk": [(1, 1, True)]}, "/trunk", 1
            )
        )

    def test_includes_range(self):
        self.assertTrue(
            properties.mergeinfo_includes_revision(
                {"/trunk": [(1, 5, True)]}, "/trunk", 3
            )
        )

    def test_includes_invalid_path(self):
        self.assertFalse(
            properties.mergeinfo_includes_revision(
                {"/somepath": [(1, 5, True)]}, "/trunk", 3
            )
        )

    def test_includes_invalid_revnum(self):
        self.assertFalse(
            properties.mergeinfo_includes_revision(
                {"/trunk": [(1, 5, True)]}, "/trunk", 30
            )
        )


class MergeInfoAddRevisionTests(TestCase):
    def test_add_new_path(self):
        mergeinfo = {}
        result = properties.mergeinfo_add_revision(mergeinfo, "/trunk", 5)
        self.assertEqual({"/trunk": [(5, 5, True)]}, result)

    def test_add_to_existing_path(self):
        mergeinfo = {"/trunk": [(1, 3, True)]}
        result = properties.mergeinfo_add_revision(mergeinfo, "/trunk", 5)
        self.assertEqual({"/trunk": [(1, 3, True), (5, 5, True)]}, result)

    def test_add_extends_range(self):
        mergeinfo = {"/trunk": [(1, 3, True)]}
        result = properties.mergeinfo_add_revision(mergeinfo, "/trunk", 4)
        self.assertEqual({"/trunk": [(1, 4, True)]}, result)

    def test_add_already_included(self):
        mergeinfo = {"/trunk": [(1, 5, True)]}
        result = properties.mergeinfo_add_revision(mergeinfo, "/trunk", 3)
        self.assertEqual({"/trunk": [(1, 5, True)]}, result)


class IsValidPropertyNameTests(TestCase):
    def test_simple_name(self):
        self.assertTrue(properties.is_valid_property_name("svn:log"))

    def test_name_with_colon(self):
        self.assertTrue(properties.is_valid_property_name(":foo"))

    def test_name_with_underscore_start(self):
        self.assertTrue(properties.is_valid_property_name("_foo"))

    def test_alphanumeric(self):
        self.assertTrue(properties.is_valid_property_name("abc123"))

    def test_with_dash(self):
        self.assertTrue(properties.is_valid_property_name("my-prop"))

    def test_with_dot(self):
        self.assertTrue(properties.is_valid_property_name("my.prop"))

    def test_invalid_start(self):
        self.assertFalse(properties.is_valid_property_name("-foo"))

    def test_invalid_char(self):
        self.assertFalse(properties.is_valid_property_name("foo bar"))

    def test_svn_prefix(self):
        self.assertTrue(properties.is_valid_property_name("svn:externals"))


class PropertyDiffTests(TestCase):
    def test_diff_empty(self):
        self.assertEqual({}, properties.diff({}, {}))

    def test_diff_added(self):
        self.assertEqual({"key": (None, "val")}, properties.diff({"key": "val"}, {}))

    def test_diff_changed(self):
        self.assertEqual(
            {"key": ("old", "new")}, properties.diff({"key": "new"}, {"key": "old"})
        )

    def test_diff_unchanged(self):
        self.assertEqual({}, properties.diff({"key": "same"}, {"key": "same"}))

    def test_diff_multiple_changes(self):
        result = properties.diff(
            {"a": "1", "b": "changed", "c": "3"}, {"a": "1", "b": "original"}
        )
        self.assertEqual({"b": ("original", "changed"), "c": (None, "3")}, result)

    def test_diff_only_reports_current_keys(self):
        # diff() only iterates over current.items(), so properties
        # that were deleted (in previous but not current) are not reported
        result = properties.diff({}, {"deleted": "val"})
        self.assertEqual({}, result)


class PropertyConstantsTests(TestCase):
    def test_prop_executable(self):
        self.assertEqual(properties.PROP_EXECUTABLE, "svn:executable")

    def test_prop_externals(self):
        self.assertEqual(properties.PROP_EXTERNALS, "svn:externals")

    def test_prop_mergeinfo(self):
        self.assertEqual(properties.PROP_MERGEINFO, "svn:mergeinfo")

    def test_prop_revision_log(self):
        self.assertEqual(properties.PROP_REVISION_LOG, "svn:log")

    def test_prop_revision_author(self):
        self.assertEqual(properties.PROP_REVISION_AUTHOR, "svn:author")

    def test_prop_revision_date(self):
        self.assertEqual(properties.PROP_REVISION_DATE, "svn:date")

    def test_prop_special(self):
        self.assertEqual(properties.PROP_SPECIAL, "svn:special")

    def test_prop_prefix(self):
        self.assertEqual(properties.PROP_PREFIX, "svn:")


class ExternalsParserAdditionalTests(TestCase):
    def test_parse_swapped_with_revision_dash_r_x(self):
        self.assertEqual(
            {"ext": (10, "http://example.com/foo")},
            properties.parse_externals_description(
                "http://example.com", "-r10 http://example.com/foo ext"
            ),
        )

    def test_parse_dir_dash_r_x_url(self):
        self.assertEqual(
            {"ext": (10, "http://example.com/foo")},
            properties.parse_externals_description(
                "http://example.com", "ext -r10 http://example.com/foo"
            ),
        )

    def test_parse_empty(self):
        self.assertEqual(
            {}, properties.parse_externals_description("http://example.com", "")
        )

    def test_parse_comment_only(self):
        self.assertEqual(
            {},
            properties.parse_externals_description(
                "http://example.com", "# just a comment"
            ),
        )


class MergeInfoPropertyCreatorAdditionalTests(TestCase):
    def test_uninheritable_range(self):
        self.assertEqual(
            "/trunk:1-2*\n",
            properties.generate_mergeinfo_property({"/trunk": [(1, 2, False)]}),
        )

    def test_uninheritable_individual(self):
        self.assertEqual(
            "/trunk:1*\n",
            properties.generate_mergeinfo_property({"/trunk": [(1, 1, False)]}),
        )

    def test_multiple_ranges(self):
        result = properties.generate_mergeinfo_property(
            {"/trunk": [(1, 3, True), (5, 8, True)]}
        )
        self.assertEqual("/trunk:1-3,5-8\n", result)
